/*
   (c) Copyright 2000-2002  convergence integrated media GmbH.
   (c) Copyright 2002       convergence GmbH.
   
   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>,
              Andreas Hundt <andi@fischlustig.de> and
              Sven Neumann <sven@convergence.de>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>


#include <directfb.h>
#include <directfb_internals.h>

#include <display/idirectfbsurface.h>

#include <media/idirectfbimageprovider.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/layers.h>
#include <core/surfaces.h>

#include <misc/gfx_util.h>
#include <misc/mem.h>
#include <misc/util.h>

#include <netinet/in.h>

static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx );

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           IDirectFBDataBuffer    *buffer );

#include <interface_implementation.h>

DFB_INTERFACE_IMPLEMENTATION( IDirectFBImageProvider, BMP )


/*
 * private data struct of IDirectFBImageProvider_BMP
 */
typedef struct {
     int                  ref;      /* reference counter */

     IDirectFBDataBuffer *buffer;
     
     
     unsigned int  Width;
     unsigned int  Height;
     __u8          ColorMap[3][256];
     unsigned int  BitPixel;
     unsigned int  ColorResolution;
     int rshift,gshift,bshift;
     int rbits,gbits,bbits;
     unsigned int rmask,gmask,bmask;

} IDirectFBImageProvider_BMP_data;

static bool ZeroDataBlock = false;

static __u32* ReadBMP( IDirectFBImageProvider_BMP_data *data, int imageNumber,
                       int *width, int *height, bool *transparency,
                       __u32 *key_rgb, bool alpha, bool headeronly);

static bool ReadOK( IDirectFBDataBuffer *buffer, void *data, unsigned int len );


static DFBResult
IDirectFBImageProvider_BMP_AddRef  ( IDirectFBImageProvider *thiz );

static DFBResult
IDirectFBImageProvider_BMP_Release ( IDirectFBImageProvider *thiz );

static DFBResult
IDirectFBImageProvider_BMP_RenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *destination_rect );

static DFBResult
IDirectFBImageProvider_BMP_SetRenderCallback( IDirectFBImageProvider *thiz,
                                              DIRenderCallback        callback,
                                              void                   *context );

static DFBResult
IDirectFBImageProvider_BMP_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                  DFBSurfaceDescription  *dsc );

static DFBResult
IDirectFBImageProvider_BMP_GetImageDescription( IDirectFBImageProvider *thiz,
                                                DFBImageDescription    *dsc );


static DFBResult
Probe( IDirectFBImageProvider_ProbeContext *ctx )
{
     if (strncmp (ctx->header, "BM", 2) == 0)
          return DFB_OK;

     return DFB_UNSUPPORTED;
}

static DFBResult
Construct( IDirectFBImageProvider *thiz,
           IDirectFBDataBuffer    *buffer )
{
     DFB_ALLOCATE_INTERFACE_DATA(thiz, IDirectFBImageProvider_BMP)

     data->ref    = 1;
     data->buffer = buffer;

     
     buffer->AddRef( buffer );

     thiz->AddRef = IDirectFBImageProvider_BMP_AddRef;
     thiz->Release = IDirectFBImageProvider_BMP_Release;
     thiz->RenderTo = IDirectFBImageProvider_BMP_RenderTo;
     thiz->SetRenderCallback = IDirectFBImageProvider_BMP_SetRenderCallback;
     thiz->GetImageDescription = IDirectFBImageProvider_BMP_GetImageDescription;
     thiz->GetSurfaceDescription =
                               IDirectFBImageProvider_BMP_GetSurfaceDescription;

     return DFB_OK;
}

static void
IDirectFBImageProvider_BMP_Destruct( IDirectFBImageProvider *thiz )
{
     IDirectFBImageProvider_BMP_data *data =
                                   (IDirectFBImageProvider_BMP_data*)thiz->priv;

     data->buffer->Release( data->buffer );

     DFB_DEALLOCATE_INTERFACE( thiz );
}

static DFBResult
IDirectFBImageProvider_BMP_AddRef( IDirectFBImageProvider *thiz )
{
     INTERFACE_GET_DATA(IDirectFBImageProvider_BMP)

     data->ref++;

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_BMP_Release( IDirectFBImageProvider *thiz )
{
     INTERFACE_GET_DATA(IDirectFBImageProvider_BMP)

     if (--data->ref == 0) {
          IDirectFBImageProvider_BMP_Destruct( thiz );
     }

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_BMP_RenderTo( IDirectFBImageProvider *thiz,
                                     IDirectFBSurface       *destination,
                                     const DFBRectangle     *dest_rect )
{
     DFBRectangle           rect = { 0, 0, 0, 0 };
     DFBSurfacePixelFormat  format;
     IDirectFBSurface_data *dst_data;
     CoreSurface           *dst_surface;
     int err;

     INTERFACE_GET_DATA (IDirectFBImageProvider_BMP)

     dst_data = (IDirectFBSurface_data*) destination->priv;
     if (!dst_data)
          return DFB_DEAD;

     dst_surface = dst_data->surface;
     if (!dst_surface)
          return DFB_DESTROYED;

     err = destination->GetSize( destination, &rect.w, &rect.h );
     if (err)
          return err;

     err = destination->GetPixelFormat( destination, &format );
     if (err)
          return err;

     /* actual loading and rendering */
     if (dest_rect == NULL || dfb_rectangle_intersect ( &rect, dest_rect )) {
          __u32 *image_data;
          bool  transparency;
          void  *dst;
          int    pitch, src_width, src_height;

          image_data = ReadBMP( data, 1, &src_width, &src_height,
                                &transparency, NULL,
                                DFB_PIXELFORMAT_HAS_ALPHA (format),
                                false );

          if (image_data) {
               err = destination->Lock( destination, DSLF_WRITE, &dst, &pitch );
               if (err) {
                    DFBFREE( image_data );
                    return err;
               }

               dst += rect.x * DFB_BYTES_PER_PIXEL(format) + rect.y * pitch;

               dfb_scale_linear_32( dst, image_data,
                                    src_width, src_height, rect.w, rect.h,
                                    pitch, format, dst_surface->palette );

               destination->Unlock( destination );
               DFBFREE(image_data);
          }
     }

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_BMP_SetRenderCallback( IDirectFBImageProvider *thiz,
                                              DIRenderCallback        callback,
                                              void                   *context )
{
     return DFB_UNIMPLEMENTED;
}

static DFBResult
IDirectFBImageProvider_BMP_GetSurfaceDescription( IDirectFBImageProvider *thiz,
                                                  DFBSurfaceDescription  *dsc )
{
     int  width;
     int  height;
     bool transparency;
     
     INTERFACE_GET_DATA (IDirectFBImageProvider_BMP)
          
     ReadBMP( data, 1, &width, &height,
              &transparency, NULL, false, true );

     dsc->flags       = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
     dsc->width       = width;
     dsc->height      = height;
     dsc->pixelformat = dfb_primary_layer_pixelformat();

     return DFB_OK;
}

static DFBResult
IDirectFBImageProvider_BMP_GetImageDescription( IDirectFBImageProvider *thiz,
                                                DFBImageDescription    *dsc )
{
     int   width;
     int   height;
     bool  transparency;
     __u32 key_rgb;
     
     INTERFACE_GET_DATA (IDirectFBImageProvider_BMP)

     ReadBMP( data, 1, &width, &height,
              &transparency, &key_rgb, false, true );

     if (transparency) {
          dsc->caps = DICAPS_COLORKEY;

          dsc->colorkey_r = (key_rgb & 0xff0000) >> 16;
          dsc->colorkey_g = (key_rgb & 0x00ff00) >>  8;
          dsc->colorkey_b = (key_rgb & 0x0000ff);
     }
     else
          dsc->caps = DICAPS_NONE;

     return DFB_OK;
}

void bs_2(unsigned short *y)
{
	unsigned char tmp,*x = (unsigned char *)y;
	tmp = x[0];
	x[0] = x[1];
	x[1] = tmp;
}

void bs_4(unsigned int *y)
{
	unsigned char tmp, *x = (unsigned char *)y;
	tmp = x[0];
	x[0] = x[3];
	x[3] = tmp;
	tmp = x[1];
	x[1] = x[2];
	x[2] = tmp;
}


typedef struct {
	   unsigned short int type;                 /* Magic identifier            */
	   unsigned int size;                       /* File size in bytes          */
	   unsigned short int reserved1, reserved2;
	   unsigned int offset;                     /* Offset to image data, bytes */
} HEADER;

void bs_HEADER(HEADER *h)
{
	// bs_2(&h->type);
	bs_4(&h->size);
	bs_4(&h->offset);
}

typedef struct {
	   unsigned int size;               /* Header size in bytes      */
	   int width,height;                /* Width and height of image */
	   unsigned short int planes;       /* Number of colour planes   */
	   unsigned short int bits;         /* Bits per pixel            */
	   unsigned int compression;        /* Compression type          */
	   unsigned int imagesize;          /* Image size in bytes       */
	   int xresolution,yresolution;     /* Pixels per meter          */
	   unsigned int ncolours;           /* Number of colours         */
	   unsigned int importantcolours;   /* Important colours         */

	   unsigned int rmask,gmask,bmask,tmask;
	   unsigned int dummy[32]; 
} INFOHEADER;


void bs_INFOHEADER(INFOHEADER *h)
{
	bs_4(&h->width);
	bs_4(&h->height);
	bs_2(&h->planes);
	bs_2(&h->bits);
	bs_4(&h->compression);
	bs_4(&h->imagesize);
	bs_4(&h->xresolution);
	bs_4(&h->yresolution);
	bs_4(&h->ncolours);
	bs_4(&h->importantcolours);
	bs_4(&h->rmask);
	bs_4(&h->gmask);
	bs_4(&h->bmask);
	bs_4(&h->tmask);
}

static int count_zeroes(unsigned int x)
{
	int i=0;
	while( (x & 1) == 0 && i<32) {
		i++;
		x>>=1;
	}
	return i;
}
static void get_pixel_16(IDirectFBImageProvider_BMP_data *data,
			     unsigned char *src,
			     unsigned char *dst)
{
	unsigned int pixel;
	unsigned int val;
	pixel = src[0] | (src[1]<<8);
	dst[0] = 0xff;
	val = ((pixel >> data->rshift) & data->rmask );
        dst[1] = (val*255) / data->rmask;  
	val = ((pixel >> data->gshift) & data->gmask );
        dst[2] = (val*255) / data->gmask;  
	val = ((pixel >> data->bshift) & data->bmask );
        dst[3] = (val*255) / data->bmask;  
}

static void get_pixel_24(IDirectFBImageProvider_BMP_data *data,
			     unsigned char *src,
			     unsigned char *dst)
{
	unsigned int pixel;
	unsigned int val;
	pixel = src[0] | (src[1]<<8) | (src[2]<<16);
	dst[0] = 0xff;
	val = ((pixel >> data->rshift) & data->rmask );
        dst[1] = (val*255) / data->rmask;  
	val = ((pixel >> data->gshift) & data->gmask );
        dst[2] = (val*255) / data->gmask;  
	val = ((pixel >> data->bshift) & data->bmask );
        dst[3] = (val*255) / data->bmask;  
}

static __u32* 
ReadBMP( IDirectFBImageProvider_BMP_data *data, int imageNumber,
                       int *width, int *height, bool *transparency,
                       __u32 *key_rgb, bool alpha, bool headeronly)
{
     DFBResult ret;
     HEADER h;
     INFOHEADER ih;
     unsigned int *image_data;
     unsigned int palette[256];

     /* FIXME: support streamed buffers */
     ret = data->buffer->SeekTo( data->buffer, 0 );
     if (ret) {
          DirectFBError( "(DirectFB/ImageProvider_BMP) Unable to seek", ret );
          return NULL;
     }

     if (! ReadOK( data->buffer, &h.type, 2  )) {
	     fprintf(stderr,"can't read header\n");
     }
     if(h.type != 'BM') {
	     fprintf(stderr,"not BMP\n");
	     return NULL;
     }
     if (! ReadOK( data->buffer, &h.size, 12  )) {
	     fprintf(stderr,"can't read header\n");
	     return NULL;
     }
     bs_HEADER(&h);
     fprintf(stderr,"sizeof(ih) = %d\n",sizeof(ih) );
     if (! ReadOK( data->buffer, &ih.size, 4  )) {
	     fprintf(stderr,"can't read header\n");
	     return NULL;
     }
     bs_4(&ih.size);
     if(ih.size < 40 || ih.size > 124) {
	     fprintf(stderr,"invalid info header size\n");
	     return NULL;
     }
     if (! ReadOK( data->buffer, &ih.width, ih.size-4 )) {
	     fprintf(stderr,"can't read header\n");
	     return NULL;
     }
     bs_INFOHEADER(&ih);
     fprintf(stderr,"ih.size = %d compression=%d\n",ih.size,
		     ih.compression);
     if(ih.size==40 || ih.compression!=3) {
	     switch(ih.bits) {
		     case 16:
			     ih.bmask=0x1f;
			     ih.gmask=0x1f<<5;
			     ih.rmask=0x1f<<10;
			     ih.compression=3;
			     break;
		     case 24:
		     case 32:
			     ih.bmask=0xff;
			     ih.gmask=0xff<<8;
			     ih.rmask=0xff<<16;
			     ih.compression=3;
			     break;
	     }
     }
     fprintf(stderr,"size=%08x offset=%08x w=%08x h=%08x bpp=%d\n",h.size,h.offset,ih.width,ih.height,
		     ih.bits);
     switch(ih.compression) {
	     case 0:
		     break;
	     case 3:
		     data->rshift = count_zeroes(ih.rmask);
		     data->gshift = count_zeroes(ih.gmask);
		     data->bshift = count_zeroes(ih.bmask);
		     data->rmask = ih.rmask >> data->rshift;
		     data->gmask = ih.gmask >> data->gshift;
		     data->bmask = ih.bmask >> data->bshift;
		     data->rbits = count_zeroes(~data->rmask);
		     data->gbits = count_zeroes(~data->gmask);
		     data->bbits = count_zeroes(~data->bmask);
		     fprintf(stderr,"shift=%d %d %d\n",data->rshift,data->gshift,data->bshift);
		     fprintf(stderr,"mask=%x %x %x\n",data->rmask,data->gmask,data->bmask);
		     fprintf(stderr,"bits=%x %x %x\n",data->rbits,data->gbits,data->bbits);
		     break;
	     default:
		     fprintf(stderr,"compression=%d not supported\n",ih.compression);
		     return NULL;
     }

     *width = ih.width;
     *height = ih.height;
     if(headeronly)
	     return NULL;
     if(ih.bits<=8) {
	     int i;
	     fprintf(stderr,"ih.ncolours=%d\n",ih.ncolours);
	     for(i=0;i<1<<ih.bits;i++) {
		 unsigned char buf[4];
		 ReadOK(data->buffer,buf,4);
		 palette[i] = 0xff000000 | (buf[0]) | (buf[1]<<8) | (buf[2]<<16);
	     }
#if 0
	     if(!ih.ncolours) {
		     int j;
		     switch(ih.bits) {
			     case 1:
				     palette[0] = 0xff000000;
				     palette[1] = 0xffffffff;
				     break;
			     case 4:
				     for(i=0;i<16;i++) {
					     palette[i] = i;
					     for(j=0;j<5;j++)
						     palette[i] = (palette[i]<<4) | palette[i];
					     palette[i] |= 0xff000000;
				     }
				     break;
			     case 8:
				     for(i=0;i<256;i++) {
					     palette[i] = 0xff |
						     (i<<16) | (i<<8) | i;
				     }
				     break;
		     }
	     }
#endif

     }
     ret = data->buffer->SeekTo( data->buffer, h.offset );

     if(ih.bits == 8 || ih.bits==4 || ih.bits==1) {
	     int x,y,i;
	     unsigned int *p;
	     unsigned char buf[4096];
	     image_data = malloc(ih.width * ih.height * 4);
	     if(!image_data) {
		     fprintf(stderr,"BMP: out of mem\n");
		     return NULL;
	     }
	     for(y=0;y<ih.height;y++) {
		     p = image_data + (ih.height-1-y)*ih.width;
		     ReadOK(data->buffer,buf,
				     ((ih.bits*ih.width+31) & ~31)/8);
		     if(ih.bits==8) 
			     for(x=0;x<ih.width;x++) 
				     *p++ = palette[buf[x]];
		     if(ih.bits==4) {
			     for(x=0;x<ih.width/2;x++) {
				     *p++ = palette[buf[x] & 0xf];
				     *p++ = palette[(buf[x] & 0xf0)>>4];
			     }
			     if(ih.width & 1)
				     *p++ = palette[buf[x] & 0xf];
		     }
		     if(ih.bits==1) {
			     for(x=0;x<ih.width/8;x++) {
				     for(i=128;i;i>>=1) {
					     if(buf[x] & i)
						     *p++ = palette[1];
					     else
						     *p++ = palette[0];
				     }
			     }
			     if(ih.width & 7) {
				     int n = 128>>(ih.width & 7);
				     for(i=128;i!=n;i>>=1) {
					     if(buf[x] & i)
						     *p++ = palette[1];
					     else
						     *p++ = palette[0];
				     }
			     }
		     }
	     }
	     return image_data;
     }
     if(ih.bits > 8 ) {
	     int x,y,nbytes;
	     unsigned char *p;
	     unsigned char buf[3*4096],*src;
	     void (*get_pixel)(IDirectFBImageProvider_BMP_data *data,
			     unsigned char *src,
			     unsigned char *dst);
	     switch(ih.bits) {
		     case 16:
			     nbytes=2;
			     get_pixel=get_pixel_16;
			     break;
		     case 24:
			     nbytes=3;
			     get_pixel=get_pixel_24;
			     break;
		     case 32:
			     nbytes=4;
			     get_pixel=get_pixel_24;
			     break;
		     default:
			     fprintf(stderr,"ih.bits=%d not supported\n",ih.bits);
			     return NULL;
	     }
		     
	     image_data = malloc(ih.width * ih.height * 4);
	     if(!image_data) {
		     fprintf(stderr,"BMP: out of mem\n");
		     return NULL;
	     }

	     for(y=0;y<ih.height;y++) {
		     int n;
		     p = image_data + (ih.height-1-y)*ih.width;
		     src = buf;
		     ReadOK(data->buffer,buf,
				     ((ih.bits*ih.width+31) & ~31)/8);
		     for(x=0;x<ih.width;x++,p+=4,src+=nbytes)
			     get_pixel(data,src,p);

	     }
	     return image_data;
     }

     return NULL;
}

static bool
ReadOK( IDirectFBDataBuffer *buffer, void *data, unsigned int len )
{
     DFBResult ret;

     ret = buffer->WaitForData( buffer, len );
     if (ret) {
          DirectFBError( "(DirectFB/ImageProvider_BMP) WaitForData failed", ret );
          return false;
     }

     ret = buffer->GetData( buffer, len, data, NULL );
     if (ret) {
          DirectFBError( "(DirectFB/ImageProvider_BMP) GetData failed", ret );
          return false;
     }

     return true;
}

