#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <linux/posix_types.h>
#undef dev_t
#define dev_t __kernel_dev_t
#include <linux/loop.h>
#undef dev_t

typedef struct {
  int bytes_per_sector;
  int sectors_per_cluster;
  int reserved_sectors;
  int number_of_fats;
  int root_entries;
  int fs_info_sector;
  int total_sectors;
  int root_dir_sectors;
  int data_sectors;
  int sectors_per_fat;
  int clusters_count;
  int type;
} FatInfo;

int le_get16(unsigned char *x)
{
	return x[0] | (x[1]<<8);
}

int le_get32(unsigned char *x)
{
	return x[0] | (x[1]<<8) | (x[2]<<16) | (x[3]<<24);
}

static int progress2_max = 100;
static int progress2_i = 0;
static inline void progress2()
{
	progress2_i++;
	if(progress2_i == progress2_max) {
		progress2_i = 0;
		write(1,".",1);
	}
}


int read_fat_info(FILE *f,FatInfo *info)
{
  unsigned char buf[4096];

  fread(buf,4096,1,f);
  
  info->bytes_per_sector = le_get16(buf+11);
  info->sectors_per_cluster = buf[13];
  info->reserved_sectors = le_get16(buf+14);
  info->number_of_fats = buf[16];
  info->root_entries = le_get16(buf+17);
  info->fs_info_sector = le_get16(buf+48);
  info->total_sectors = le_get32(buf+32);
  if(!info->total_sectors)
	  info->total_sectors = le_get16(buf+19);
  info->total_sectors--;
  // Count of sectors occupied by one FAT
  info->sectors_per_fat = le_get32(buf+36);
  if(!info->sectors_per_fat)
	  info->sectors_per_fat = le_get16(buf+22);
  
  // calculate the FAT Type with the cluster count (FAT12 or FAT16 or FAT32)
  info->root_dir_sectors = ((info->root_entries * 32) + (info->bytes_per_sector-1)) / info->bytes_per_sector;
  
  info->data_sectors = info->total_sectors - (info->reserved_sectors + (info->number_of_fats * info->sectors_per_fat)
		  + info->root_dir_sectors);
  
  if (info->sectors_per_cluster == 0)
    {
	    return -1;
    }
  
  info->clusters_count = info->data_sectors / info->sectors_per_cluster;
  progress2_max = info->clusters_count/100;
  if(info->clusters_count < 4085)
	  info->type = 12;
  else if(info->clusters_count < 65525)
	  info->type = 16;
  else
	  info->type = 32;

  return 0;  
}

int find_max_free_region(FILE *f,FatInfo *info,int *minp,int *maxp,int *used_p)
{
	int reserved_bytes;
	int i;
	int min,max,min2,max2;
	int count;
	int free_clusters=0,used_clusters=0;
	unsigned int entry;
  	reserved_bytes = info->reserved_sectors * info->bytes_per_sector;
	if(info->type != 32) {
		fprintf(stderr,"must be FAT32\n");
		return -1;
	}
	fseek(f,reserved_bytes+2*4,SEEK_SET);
	min = -1;
	max = -1;
	min2 = -1;
	max2 = -1;
	count = info->clusters_count;
	
  
  	for (i=0; i < count ; i++) {
		fread(&entry,4,1,f);
		if(entry==0) {
			free_clusters++;
			if(min < 0)
				min = i;
			max = i;
			if(max-min > max2-min2) {
				max2=max;
				min2=min;
			}
		} else {
			used_clusters++;
			min = -1;
			max = -1;
		}
		progress2();
	}
	fprintf(stderr,"min2=%d max2=%d free=%d used=%d\n",min2,max2,free_clusters,used_clusters);
	*minp = min2;
	*maxp = max2;
	*used_p = used_clusters;
	return 0;
}

int
set_loop (const char *device, const char *file, long long offset)
{
	struct loop_info loopinfo;
	int fd, ffd, mode, i;
	char *pass;

	mode = O_RDWR;
	if ((ffd = open (file, mode)) < 0) {
		if (ffd < 0) {
			perror (file);
			return 1;
		}
	}
	if ((fd = open (device, mode)) < 0) {
		perror (device);
		return 1;
	}

	memset (&loopinfo, 0, sizeof (loopinfo));
	strncpy (loopinfo.lo_name, file, LO_NAME_SIZE);
	loopinfo.lo_offset = offset;
	printf("sizeof=%d\n",sizeof(loopinfo.lo_offset));

	loopinfo.lo_encrypt_key_size = 0;
	if (ioctl (fd, LOOP_SET_FD, ffd) < 0) {
		perror ("ioctl: LOOP_SET_FD");
		return 1;
	}
	if (ioctl (fd, LOOP_SET_STATUS, &loopinfo) < 0) {
		(void) ioctl (fd, LOOP_CLR_FD, 0);
		perror ("ioctl: LOOP_SET_STATUS");
		return 1;
	}
	close (fd);
	close (ffd);
	return 0;
}

#define PR(x,y) fprintf(stderr,"info." y "=%d\n",info.x);

long long fat32_get_free_region(char *device,int minsize,int *used)
{
	FILE *f;
	FatInfo info;
	int min,max;
	long long size;
	long long offset;
	char cmd[256];
	*used = -1;
	f = fopen(device,"r");
	if(f==NULL) {
		perror("open");
		return -1;
	}
	if( read_fat_info(f,&info) != 0) {
		fprintf(stderr,"error in read_fat_info\n");
		fclose(f);
		return -1;
	}
	PR(sectors_per_cluster,"sectors_per_cluster");
	PR( reserved_sectors,"reserved_sectors");
	PR( number_of_fats,"number_of_fats");
	PR( root_entries,"root_entries");
	PR( total_sectors,"total_sectors");
	PR( root_dir_sectors,"root_dir_sectors");
	PR( data_sectors,"data_sectors");
	PR( sectors_per_fat,"sectors_per_fat");
	PR( clusters_count,"clusters_count");
	PR( type,"type");
	find_max_free_region(f,&info,&min,&max,used);
	fclose(f);
	size = max-min;
	min+=size/4;
	max-=size/4;
	size = max-min;
	size *= info.sectors_per_cluster*info.bytes_per_sector;

	offset = info.total_sectors - info.data_sectors;
	offset += min*info.sectors_per_cluster;
	offset *= info.bytes_per_sector;
	fprintf(stderr,"offset=%lld size=%lld\n",offset,size);
	if(size < minsize) {
		fprintf(stderr,"too small\n");
		return -1;
	}
	return offset;
}
