/*
 * linux/arch/mips/vr41xx/int-handler.S
 *
 * Copyright (C) 1995, 1996, 1997 Paul M. Antoine and Harald Koerfgen
 *
 * Written by Ralf Baechle and Andreas Busse, modified for DECStation
 * support by Paul Antoine and Harald Koerfgen.
 *
 * completly rewritten:
 * Copyright (C) 1998 Harald Koerfgen
 *
 * Adapted to the VR41XX and almost entirely rewritten:
 * Copyright (C) 1999 Bradley D. LaRonde and Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/mipsregs.h>
#include <asm/stackframe.h>
#include <asm/vr41xx.h>


		.text
		.set	noreorder

/*
 * Interrupt handler for NEC uPD61130X  
 *
 * In a zealous attempt to save CPU cycles, we don't worry about making the code
 * readable and instead optimize for the most frequently occurring ints (the
 * timer int IRQ 3 and the ICU ints that cascade off IRQ 2), while keeping a
 * priority scheme that actually makes some sense:
 *
 * Highest ----     CPU IP3 - Int1: RTClong1    IRQ  3
 *                  CPU IP4 - Int2: RTClong2    IRQ  4
 *                  CPU IP5 - Int3: HSP         IRQ  5
 *                  CPU IP6 - Int4: unused      IRQ  6
 *                  CPU IP7 - counter           IRQ  7
 *                  ICU SYSINT1REG bit 0        IRQ  8
 *                  ...
 *                  ICU SYSINT1REG bit 31       IRQ 39
 *                  ICU SYSINT2REG bit 0        IRQ 40
 *                  ...
 *                  ICU SYSINT1REG bit 31       IRQ 71
 *                  CPU IP0 - software 1        IRQ  0
 * Lowest ---       CPU IP1 - software 2        IRQ  1
 *
 * upper 16-bits of SYSINTREGS still not implemented
 *
 * IRQ 2 is the cascade to the ICU (CPU IP2, Int0) Host Interface
 *
 * then we just return, if multiple IRQs are pending then we will just take
 * another exception immediately.
 *
 */
		.align	5
		NESTED(vr41xx_handle_irq, PT_SIZE, ra)
		.set	noat
		SAVE_ALL
		CLI
		.set	at
		.set	noreorder

		// Get the pending interrupts
		mfc0	t0,CP0_CAUSE

		// Isolate the allowed ones by anding with irq mask
		mfc0	t2,CP0_STATUS

		andi	t0,0xff00
		and	t0,t2
		andi	t1,t0,0x0800	// check for IRQ 3
		bnez	t1,handle_it
		 li	a0,3
		andi	t1,t0,0xf000	// check for IRQ 4-7
		bnez	t1,1f
		 andi	t1,t0,0x0400	// check for IRQ 2
		bnez	t1,icu_int
		 lui	t3,%hi(VR41XX_SYSINT1REG)
		beqz	t0,go_spurious	// check for IRQ 0-1
		 li	a0,-1
		sll	t0,4
1:		andi	t1,t0,0x1000
		srl	t0,1
		beqz	t1,1b
		 addiu	a0,1
		b	handle_it
		 nop

/*
 *  In the interest of speed, we implement a 3 tier check for which bit
 *  in the ICU level 1 interrupt registers is active.  First, we determine
 *  which 32-bit register (1 or 2).  Second, we narrow it down to which
 *  group of 4 bits, then we get the exact bit.  a0 holds the IRQ number.
 *  Since we always add both 4 and 1 to a0, we initalize it to 5 less than
 *  the lowest IRQ number for each ICU level 1 register
 */
icu_int:	// see delay slot instruction from branch that got us here
		lw	t0,%lo(VR41XX_SYSINT2REG)(t3)
		lw	t2,%lo(VR41XX_MSYSINT2REG)(t3)
		and	t0,t2
		beqz	t0,2f
		li	a0,35
		b	1f
		nop

go_spurious:	j	spurious_interrupt
		 nop

2:		lw	t0,%lo(VR41XX_SYSINT1REG)(t3)
		lw	t2,%lo(VR41XX_MSYSINT1REG)(t3)
		and	t0,t2
		beqz	t0,go_spurious
		li	a0,3

1:		andi	t1,t0,0x000f
		 addi	a0,4
		beqzl	t1,1b		// the beqzl will cancel the last
		 srl	t0,4		//  shift so we don't lose the bits

1:		andi	t1,t0,0x0001
		 addi	a0,1
		beqz	t1,1b
		 srl	t0,1

handle_it:	jal	do_IRQ
		 move	a1,sp

		j	ret_from_irq
		 nop
		
		END(vr41xx_handle_irq)
