/* $Id: dsiu_aux.c,v 1.1.1.1 2001/05/02 13:47:08 mfloeer Exp $
 *
 * Aux (PS/2 mouse) driver for VR41xx debug serial port (DSIU)
 *
 * UI and some other parts swiped from drivers/char/pc_kbd.c
 *
 * From the user perspective, this driver should behave mostly the
 * same as the aux mouse driver in pc_kbd.c with the following exception:
 * a write operation will return an EAGAIN error instead of blocking
 * if the device is opened as non-blocking and a write cannot be processed
 * immediately.  Since the write buffer is currently only 1 char, this may
 * cause software compatibility problems in (broken) user apps that do
 * non-blocking writes and don't check return value for EAGAIN.
 *
 * Copyright (c) 2000 Michael Klar <wyldfier@iname.com>
 */

#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>	// for the AUX command defs
#include <linux/poll.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/vr41xx.h>

/*
 *	PS/2 Auxiliary Device
 */

#define AUX_RECONNECT 170 /* scancode when ps2 device is plugged (back) in */
 
static struct aux_queue *queue;	/* Mouse data buffer. */
static int aux_count;		/* = 0 */
static int tx_count;
static char tx_pending;
static wait_queue_head_t tx_wait;
/* used when we send commands to the mouse that expect an ACK. */
static unsigned char mouse_reply_expected = 0;

static void dsiu_aux_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned short intr0reg = *VR41XX_INTR0REG;
	int head;

	if (intr0reg & 0x0002) {
		unsigned short data = *VR41XX_RXB0LREG;

		if (mouse_reply_expected) {
			if (data == AUX_ACK) {
				mouse_reply_expected--;
				goto rx_done;
			}
			mouse_reply_expected = 0;
		}

		add_mouse_randomness(data);

		head = queue->head;
		queue->buf[head] = data;
		head = (head + 1) & (AUX_BUF_SIZE-1);
		if (head != queue->tail) {
			queue->head = head;
			kill_fasync(&queue->fasync, SIGIO, POLL_IN);
			wake_up_interruptible(&queue->proc_list);
		}
rx_done:
		*VR41XX_INTR0REG = 0x0002;	/* ack */
	}
	if (intr0reg & 0x0001) {
		*VR41XX_INTR0REG = 0x0001;	/* ack */
		barrier();
		/*
		 * note: if there was a write pending completion when user
		 * opened the device, it is possible to get a tx completion
		 * int when tx_count == 0.
		 */
		if (tx_count && --tx_count)
			*VR41XX_TXS0LREG = (unsigned short)tx_pending;
		wake_up_interruptible(&tx_wait);
	}
}

/*
 * Send a byte to the mouse.
 */
static int aux_write_dev(char val, int noblock)
{
	cli();	/* to prevent tx_count from changing on us */
	while (tx_count == 2) {
		if (noblock) {
			sti();
			return -EAGAIN;
		}
		interruptible_sleep_on(&tx_wait);	/* this will sti() */
		if (signal_pending(current))
			return -ERESTARTSYS;
		cli();
	}
	if (tx_count || *VR41XX_ASIS0REG & 0x0080) {
		tx_pending = val;
		tx_count = 2;
	} else {
		*VR41XX_TXS0LREG = (unsigned short)val;
		tx_count = 1;
	}
	sti();

	return 0;
}

/*
 * Send a byte to the mouse & handle returned ack
 */
static inline void aux_write_ack(char val)
{
	while (aux_write_dev(val, 0)) ;
	/* we expect an ACK in response. */
	mouse_reply_expected++;
}

static inline unsigned char get_from_queue(void)
{
	unsigned char result;

	result = queue->buf[queue->tail];
	queue->tail = (queue->tail + 1) & (AUX_BUF_SIZE-1);
	return result;
}


static inline int queue_empty(void)
{
	return queue->head == queue->tail;
}

static int fasync_aux(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &queue->fasync);
	if (retval < 0)
		return retval;
	return 0;
}


/*
 * Random magic cookie for the aux device
 */

static int release_aux(struct inode * inode, struct file * file)
{
	fasync_aux(-1, file, 0);
	if (--aux_count)
		return 0;
	
	*VR41XX_MDSIUINTREG = 0;
	free_irq(VR41XX_IRQ_DSIU, NULL);
	return 0;
}

/*
 * Install interrupt handler.
 * Enable auxiliary device.
 */

static int open_aux(struct inode * inode, struct file * file)
{
	if (aux_count++) {
		return 0;
	}
	queue->head = queue->tail = 0;		/* Flush input queue */
	if (request_irq(VR41XX_IRQ_DSIU, dsiu_aux_interrupt, 0, "PS/2 Mouse", NULL)) {
		aux_count--;
		return -EBUSY;
	}
	tx_count = 0;
	*VR41XX_INTR0REG = 0x000f;	/* ack any pending ints */
	barrier();
	*VR41XX_MDSIUINTREG = 0x0300;	/* yes, we ignore receive errors */

	return 0;
}

/*
 * Put bytes from input queue to buffer.
 */

static ssize_t read_aux(struct file * file, char * buffer,
			size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	ssize_t i = count;
	unsigned char c;

	if (queue_empty()) {
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		add_wait_queue(&queue->proc_list, &wait);
repeat:
		set_current_state(TASK_INTERRUPTIBLE);
		if (queue_empty() && !signal_pending(current)) {
			schedule();
			goto repeat;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&queue->proc_list, &wait);
	}
	while (i > 0 && !queue_empty()) {
		c = get_from_queue();
		if (put_user(c, buffer++))
			return -EFAULT;
		i--;
	}
	if (count-i) {
		file->f_dentry->d_inode->i_atime = CURRENT_TIME;
		return count-i;
	}
	if (signal_pending(current))
		return -ERESTARTSYS;
	return 0;
}

/*
 * Write to the aux device.
 */

static ssize_t write_aux(struct file * file, const char * buffer,
			 size_t count, loff_t *ppos)
{
	ssize_t retval = 0;

	if (count) {
		ssize_t written = 0;

		if (count > 32)
			count = 32; /* Limit to 32 bytes. */
		do {
			char c;
			if (get_user(c, buffer++))
				return -EFAULT;
			retval = aux_write_dev(c, file->f_flags & O_NONBLOCK);
			if (retval)
				return retval;
			written++;
		} while (--count);
		retval = -EIO;
		if (written) {
			retval = written;
			file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
		}
	}

	return retval;
}

/* No kernel lock held - fine */
static unsigned int aux_poll(struct file *file, poll_table * wait)
{
	poll_wait(file, &queue->proc_list, wait);
	if (!queue_empty())
		return POLLIN | POLLRDNORM;
	return 0;
}

struct file_operations psaux_fops = {
	read:		read_aux,
	write:		write_aux,
	poll:		aux_poll,
	open:		open_aux,
	release:	release_aux,
	fasync:		fasync_aux,
};

/*
 * Initialize driver.
 */
static struct miscdevice psaux_mouse = {
	PSMOUSE_MINOR, "psaux", &psaux_fops
};

static int __init dsiu_aux_init(void)
{
	misc_register(&psaux_mouse);
	queue = (struct aux_queue *) kmalloc(sizeof(*queue), GFP_KERNEL);
	memset(queue, 0, sizeof(*queue));
	queue->head = queue->tail = 0;
	init_waitqueue_head(&queue->proc_list);
	init_waitqueue_head(&tx_wait);

#if 0 // MFK: need to init both the DSIU regs and the TPM754
	aux_write_ack(AUX_SET_SAMPLE);
	aux_write_ack(100);			/* 100 samples/sec */
	aux_write_ack(AUX_SET_RES);
	aux_write_ack(3);			/* 8 counts per mm */
	aux_write_ack(AUX_SET_SCALE21);		/* 2:1 scaling */
#endif
	*VR41XX_MDSIUINTREG = 0;

	return 0;
}

__initcall(dsiu_aux_init);

