/*
 *	Real Time Clock interface for VR41xx on-chip RTC
 *
 *	Based on v1.10 of Paul Gortmaker's PC RTC driver
 *	VR41xx adaptation by Michael Klar
 *
 *	Copyright (C) 1996 Paul Gortmaker
 *	Copyright (C) 2000 Michael Klar
 *
 *	This driver allows use of the real time clock (built into
 *	nearly all computers) from user space. It exports the /dev/rtc
 *	interface supporting various ioctl() and also the
 *	/proc/driver/rtc pseudo-file for status information.
 *
 *	The ioctls can be used to set the interrupt behaviour and
 *	generation rate from the RTC. Then the /dev/rtc
 *	interface can be used to make use of these timer interrupts,
 *	be they interval or alarm based.
 *
 *	The /dev/rtc interface will block on reads until an interrupt
 *	has been received. If a RTC interrupt has already happened,
 *	it will output an unsigned long and then block. The output value
 *	contains the interrupt status in the low byte and the number of
 *	interrupts since the last read in the remaining high bytes. The 
 *	/dev/rtc interface can also be used with the select(2) call.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Based on other minimal char device drivers, like Alan's
 *	watchdog, Ted's random, etc. etc.
 */

#define RTC_VERSION		"1.0"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/vr41xx.h>

/*
 *	We sponge a minor off of the misc major. No need slurping
 *	up another valuable major dev number for this. If you add
 *	an ioctl, make sure you don't conflict with SPARC's RTC
 *	ioctls.
 */

extern spinlock_t rtc_lock;
extern unsigned long get_rtc_time(volatile unsigned short *reg);
extern void set_rtc_time(unsigned long timeval, volatile unsigned short *reg);

static struct fasync_struct *rtc_async_queue;

static DECLARE_WAIT_QUEUE_HEAD(rtc_wait);

static long long rtc_llseek(struct file *file, loff_t offset, int origin);

static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos);

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);

static unsigned int rtc_poll(struct file *file, poll_table *wait);

static void get_time_regs(struct rtc_time *rtc_tm,
			  volatile unsigned short *reg);

static int set_time_regs(struct rtc_time *rtc_tm,
		         volatile unsigned short *reg);

static void mask_rtc_irq(unsigned int irq, int bit);

static int set_rtc_irq(unsigned int irq, int bit);

static int rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data);

/*
 *	Bits in rtc_status. (6 bits of room for future expansion)
 */

#define RTC_IS_OPEN		0x01	/* means /dev/rtc is in use	*/
#define RTC_TIMER_ON		0x02	/* missed irq timer active	*/
/* RTC_TIMER_ON unused, since VR41xx RTC doesn't stop int on delayed ack */
#define RTC_PIE_EN		0x04	/* keeps track of if PIE IRQ requested */
#define RTC_AIE_EN		0x08	/*  same for AIE		*/
#define RTC_IS_DST		0x10	/* daylight savings time	*/

static atomic_t rtc_status = ATOMIC_INIT(0); /* bitmapped status byte.	*/
static unsigned long rtc_freq = 0;	/* Current periodic IRQ rate	*/
static unsigned long rtc_irq_data = 0;	/* our output to the world	*/

/*
 *	If this driver ever becomes modularised, it will be really nice
 *	to make the epoch retain its value across module reload...
 */
static unsigned long epoch;	/* year corresponding to time 0		*/
extern unsigned long epoch_adj;	/* cached adjustment value, in seconds	*/

static const unsigned char days_in_mo[] = 
{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

/*
 *	A very tiny interrupt handler. It runs with SA_INTERRUPT set,
 *	so that there is no possibility of conflicting with the
 *	set_rtc_mmss() call that happens during some timer interrupts.
 *	(See ./arch/XXXX/kernel/time.c for the set_rtc_mmss() function.)
 */

static void rtc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned short intreg;

	/*
	 *	Can be an alarm interrupt, update complete interrupt,
	 *	or a periodic interrupt. We store the status in the
	 *	low byte and the number of interrupts received since
	 *	the last read in the remainder of rtc_irq_data.
	 */

	spin_lock (&rtc_lock);
	rtc_irq_data += 0x100;
	rtc_irq_data &= ~0xff;
	intreg = *VR41XX_RTCINTREG & 0x05;
	barrier();
	/* ack the interrupt(s), but only the ones we reported */
	*VR41XX_RTCINTREG = intreg;
	rtc_irq_data |= intreg;
	spin_unlock (&rtc_lock);

	wake_up_interruptible(&rtc_wait);	

	if (rtc_async_queue)
		kill_fasync (&rtc_async_queue, SIGIO, POLL_IN);
}

/*
 *	Now all the various file operations that we export.
 */

static long long rtc_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long data;
	ssize_t retval;
	
	if (count < sizeof(unsigned long))
		return -EINVAL;

	add_wait_queue(&rtc_wait, &wait);

	current->state = TASK_INTERRUPTIBLE;
		
	while ((data = xchg(&rtc_irq_data, 0)) == 0) {
		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			goto out;
		}
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto out;
		}
		schedule();
	}

	retval = put_user(data, (unsigned long *)buf); 
	if (!retval)
		retval = sizeof(unsigned long); 
 out:
	current->state = TASK_RUNNING;
	remove_wait_queue(&rtc_wait, &wait);

	return retval;
}

static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{

	struct rtc_time wtime; 

	switch (cmd) {
	case RTC_AIE_OFF:	/* Mask alarm int. enab. bit	*/
	{
		mask_rtc_irq(VR41XX_IRQ_ETIMER, RTC_AIE_EN);
		return 0;
	}
	case RTC_AIE_ON:	/* Allow alarm interrupts.	*/
	{
		return set_rtc_irq(VR41XX_IRQ_ETIMER, RTC_AIE_EN);
	}
	case RTC_PIE_OFF:	/* Mask periodic int. enab. bit	*/
	{
		mask_rtc_irq(VR41XX_IRQ_INT2, RTC_PIE_EN);
		return 0;
	}
	case RTC_PIE_ON:	/* Allow periodic ints		*/
	{

		/*
		 * We don't really want Joe User enabling more
		 * than 64Hz of interrupts on a multi-user machine.
		 */
		if ((rtc_freq > 64) && (!capable(CAP_SYS_RESOURCE)))
			return -EACCES;
		return set_rtc_irq(VR41XX_IRQ_INT2, RTC_PIE_EN);
	}
	case RTC_ALM_READ:	/* Read the present alarm time */
	{
		/*
		 * This returns a struct rtc_time. Reading >= 0xc0
		 * means "don't care" or "match all". Only the tm_hour,
		 * tm_min, and tm_sec values are filled in.
		 */

		get_time_regs(&wtime, VR41XX_ECMPLREG);
		break; 
	}
	case RTC_ALM_SET:	/* Store a time into the alarm */
	{
		/*
		 * This expects a struct rtc_time. Writing 0xff means
		 * "don't care" or "match all". Only the tm_hour,
		 * tm_min and tm_sec are used.
		 */
		struct rtc_time alm_tm;

		if (copy_from_user(&alm_tm, (struct rtc_time*)arg,
				   sizeof(struct rtc_time)))
			return -EFAULT;

		return set_time_regs(&alm_tm, VR41XX_ECMPLREG);
	}
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/
	{
		get_time_regs(&wtime, VR41XX_ETIMELREG);
		break;
	}
	case RTC_SET_TIME:	/* Set the RTC */
	{
		struct rtc_time alm_tm;
			
		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		if (copy_from_user(&alm_tm, (struct rtc_time*)arg,
				   sizeof(struct rtc_time)))
			return -EFAULT;

		return set_time_regs(&alm_tm, VR41XX_ETIMELREG);
	}
	case RTC_IRQP_READ:	/* Read the periodic IRQ rate.	*/
	{
		return put_user(rtc_freq, (unsigned long *)arg);
	}
	case RTC_IRQP_SET:	/* Set periodic IRQ rate.	*/
	{
		int quot;

		if (arg < 1)
			return -EINVAL;
		/*
		 * We don't really want Joe User generating more
		 * than 64Hz of interrupts on a multi-user machine.
		 */
		if ((arg > 64) && (!capable(CAP_SYS_RESOURCE)))
			return -EACCES;

		quot = (32768 + arg / 2) / arg;

		/* 
		 * The max we can do is 32768Hz/5 = 6553Hz.
		 */
		if (quot < 5)
			return -EINVAL;

		/*
		 * This does not ensure that the frequency requested
		 * is exactly the one requested.  Unfortunately, doing
		 * so would restrict to powers of 2, and hardware is
		 * more capable than that.  It's too bad the argument
		 * here is a frequency, not a time value.
		 */
		rtc_freq = arg;

		spin_lock_irq(&rtc_lock);
		*VR41XX_RTCL2LREG = quot;
		*VR41XX_RTCL2LREG = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	}
	case RTC_EPOCH_READ:	/* Read the epoch.	*/
	{
		return put_user (epoch, (unsigned long *)arg);
	}
	case RTC_EPOCH_SET:	/* Set the epoch.	*/
	{
		/* 
		 * There were no RTC clocks before 1970.
		 */
		if (arg < 1970)
			return -EINVAL;

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		epoch = arg;
		epoch_adj = mktime(epoch,0,0,0,0,0) - mktime(1970,0,0,0,0,0);
		return 0;
	}
	default:
		return -EINVAL;
	}
	return copy_to_user((void *)arg, &wtime, sizeof wtime) ? -EFAULT : 0;
}

/*
 *	We enforce only one user at a time here with the open/close.
 *	Also clear the previous interrupt data on an open, and clean
 *	up things on a close.
 */

static int rtc_open(struct inode *inode, struct file *file)
{
	if(atomic_read(&rtc_status) & RTC_IS_OPEN)
		return -EBUSY;

	MOD_INC_USE_COUNT;

	atomic_set(&rtc_status, atomic_read(&rtc_status) | RTC_IS_OPEN);

	spin_lock_irq(&rtc_lock);
	rtc_irq_data = 0;
	spin_unlock_irq(&rtc_lock);
	return 0;
}

static int rtc_fasync (int fd, struct file *filp, int on)

{
	return fasync_helper (fd, filp, on, &rtc_async_queue);
}

static int rtc_release(struct inode *inode, struct file *file)
{
	/*
	 * Turn off any interrupts once the device is no longer
	 * in use, and clear the data.
	 */
	mask_rtc_irq(VR41XX_IRQ_ETIMER, RTC_AIE_EN);
	mask_rtc_irq(VR41XX_IRQ_INT2, RTC_PIE_EN);

	if (file->f_flags & FASYNC) {
		rtc_fasync (-1, file, 0);
	}

	MOD_DEC_USE_COUNT;

	atomic_set(&rtc_status, atomic_read(&rtc_status) & ~RTC_IS_OPEN);
	return 0;
}

static unsigned int rtc_poll(struct file *file, poll_table *wait)
{
	unsigned long l;

	poll_wait(file, &rtc_wait, wait);

	spin_lock_irq(&rtc_lock);
	l = rtc_irq_data;
	spin_unlock_irq(&rtc_lock);

	if (l != 0)
		return POLLIN | POLLRDNORM;
	return 0;
}

/*
 *	The various file operations we support.
 */

static struct file_operations rtc_fops = {
	llseek:		rtc_llseek,
	read:		rtc_read,
	poll:		rtc_poll,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release,
	fasync:		rtc_fasync,
};

static struct miscdevice rtc_dev=
{
	RTC_MINOR,
	"rtc",
	&rtc_fops
};

static int __init rtc_init(void)
{
	misc_register(&rtc_dev);
	create_proc_read_entry ("driver/rtc", 0, 0, rtc_read_proc, NULL);

	/* Initialize periodic freq. to what PC RTC driver uses, which is 1024Hz */
	*VR41XX_RTCL2LREG = 32768 / 1024;
	*VR41XX_RTCL2LREG = 0;
	rtc_freq = 1024;
	epoch = 1970;

	printk(KERN_INFO "VR41xx Real Time Clock Driver v" RTC_VERSION "\n");

	return 0;
}

static void __exit rtc_exit (void)
{
	/* interrupts disabled at this point by rtc_release */

	epoch_adj = 0;
	remove_proc_entry ("driver/rtc", NULL);
	misc_deregister(&rtc_dev);
}

module_init(rtc_init);
module_exit(rtc_exit);
EXPORT_NO_SYMBOLS;

/*
 *	Info exported via "/proc/driver/rtc".
 */

static int rtc_proc_output (char *buf)
{
	char *p;
	struct rtc_time tm;
	int status;

	p = buf;

	get_time_regs(&tm, VR41XX_ETIMELREG);

	/*
	 * There is no way to tell if the luser has the RTC set for local
	 * time or for Universal Standard Time (GMT). Probably local though.
	 */
	p += sprintf(p,
		     "rtc_time\t: %02d:%02d:%02d\n"
		     "rtc_date\t: %04d-%02d-%02d\n"
	 	     "rtc_epoch\t: %04lu\n",
		     tm.tm_hour, tm.tm_min, tm.tm_sec,
		     tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, epoch);

	get_time_regs(&tm, VR41XX_ECMPLREG);

	/*
	 * We implicitly assume 24hr mode here.
	 */
	p += sprintf(p,
		     "alarm time\t: %02d:%02d:%02d\n"
		     "alarm date\t: %04d-%02d-%02d\n",
		     tm.tm_hour, tm.tm_min, tm.tm_sec,
		     tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);

	status = atomic_read(&rtc_status);

	p += sprintf(p,
		     "DST_enable\t: %s\n"
		     "alarm_IRQ\t: %s\n"
		     "periodic_IRQ\t: %s\n"
		     "periodic_freq\t: %ld\n",
		     (status & RTC_IS_DST) ? "yes" : "no",
		     (status & RTC_AIE_EN) ? "yes" : "no",
		     (status & RTC_PIE_EN) ? "yes" : "no",
		     rtc_freq);

	return  p - buf;
}

static int rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data)
{
        int len = rtc_proc_output (page);
        if (len <= off+count) *eof = 1;
        *start = page + off;
        len -= off;
        if (len>count) len = count;
        if (len<0) len = 0;
        return len;
}

/*
 * Used to disable/enable interrupts for any one of UIE, AIE, PIE.
 * We also clear out any old irq data after an ioctl() that
 * meddles with the interrupt enable/disable bits.
 */

static void mask_rtc_irq(unsigned int irq, int bit)
{
	spin_lock_irq(&rtc_lock);
	if (atomic_read(&rtc_status) & bit) {
		atomic_set(&rtc_status, atomic_read(&rtc_status) & ~bit);
		free_irq(irq, NULL);
	}
	rtc_irq_data = 0;
	spin_unlock_irq(&rtc_lock);
}

static int set_rtc_irq(unsigned int irq, int bit)
{
	int retval = 0;

	spin_lock_irq(&rtc_lock);
	if (!(atomic_read(&rtc_status) & bit)) {
		atomic_set(&rtc_status, atomic_read(&rtc_status) | bit);
		retval = request_irq(irq, rtc_interrupt, SA_INTERRUPT, "rtc", NULL);
	}
	rtc_irq_data = 0;
	spin_unlock_irq(&rtc_lock);

	return retval;
}

static void get_time_regs(struct rtc_time *rtc_tm, volatile unsigned short *reg)
{
	unsigned long timeval, year, days;
	long temp;
	int leap_yr = 0;

	spin_lock_irq(&rtc_lock);
	timeval = get_rtc_time(reg);	// seconds since 1970-Jan-1 00:00:00
	spin_unlock_irq(&rtc_lock);
	rtc_tm->tm_sec = timeval % 60;
	timeval /= 60;			// minutes
	rtc_tm->tm_min = timeval % 60;
	timeval /= 60;			// hours

	// adjust for daylight savings time, if enabled
	if (atomic_read(&rtc_status) & RTC_IS_DST) {
		rtc_tm->tm_isdst = 1;
		timeval++;
	} else {
		rtc_tm->tm_isdst = 0;
	}
	rtc_tm->tm_hour = timeval % 24;
	timeval /= 24;			// days
	rtc_tm->tm_wday = (timeval + 4) % 7 + 1;

	days = timeval + 719468;	// adjust for 1970, put leap day last
	year = (days / (365 * 400 + 97)) * 400;
	days = days % (365 * 400 + 97);	// day offset into 400 years
	temp = (days * 4 + 3) / (365 * 400 + 97);
	year += temp * 100;
	days -= temp * (365 * 100 + 24); // day offset into 100 years
	year += (days / (365 * 4 + 1)) * 4;
	days = days % (365 * 4 + 1);	// day offset into 4 years
	temp = (days * 4 + 3) / (4 * 365 + 1);
	year += temp;
	days -= temp * 365;		// day offset into year, 0 = Mar 1
	
	temp = (days * 5 + 2) / 153;
	rtc_tm->tm_mday = days - (temp * 153 + 2) / 5 + 1;
	if (temp < 10) {
		temp += 12;
		days += 365;
		leap_yr = ((!(year % 4) && (year % 100)) || !(year % 400));
		year -= 1;
	}
	rtc_tm->tm_mon = temp - 10;
	rtc_tm->tm_yday = days - 305 + leap_yr;
	rtc_tm->tm_year = year - 1899;
}

static int set_time_regs(struct rtc_time *rtc_tm, volatile unsigned short *reg)
{
	unsigned long timeval;
	unsigned int year, leap_day;

	year = rtc_tm->tm_year + 1900;
	leap_day = (rtc_tm->tm_mon == 1) && ((!(year % 4) && (year % 100))
	             || !(year % 400));
	if ((year < 1970) || (rtc_tm->tm_mon > 11) ||
	    (rtc_tm->tm_hour > 24) || (rtc_tm->tm_min > 60) ||
	    (rtc_tm->tm_sec > 60) || (rtc_tm->tm_mday == 0) ||
	    (rtc_tm->tm_mday > days_in_mo[rtc_tm->tm_mon] + leap_day))
		return -EINVAL;

	timeval = mktime(year, rtc_tm->tm_mon + 1, rtc_tm->tm_mday,
			 rtc_tm->tm_hour, rtc_tm->tm_min, rtc_tm->tm_sec);

	// Note that we keep this in RAM, so it does not survive reboot
	if (rtc_tm->tm_isdst) {
		atomic_set(&rtc_status, atomic_read(&rtc_status) | RTC_IS_DST);
		timeval -= 60 * 60;
	} else {
		atomic_set(&rtc_status, atomic_read(&rtc_status) & ~RTC_IS_DST);
	}
	set_rtc_time(timeval, reg);
	return 0;
}
