/*
 * linux/arch/mips/vr41xx/vrc4173-kbd.c
 *
 * VRC4173 KIU support
 *
 * Copyright (C) 1999 Bradley D. LaRonde
 * Oct 1999 Hiroshi Kawashima (kawashima@iname.com) adjusted scan lines for MobilPro.
 * Copyright (C) 2000 Michael Klar <wyldfier@iname.com>
 * Copyright (C) 2000 Michael R. McDonald <mikemac@coolcs.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The PS2 code originally came from arch/mips/vr41xx/kbd.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>
#include <linux/kbd_kern.h>
#include <linux/kbd_ll.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <asm/io.h>
#include <asm/keyboard.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#include <asm/vr41xx-platdep.h>

extern struct pci_dev *vrc4173_dev;

extern spinlock_t vrc4173_controller_lock;

void vrc4173_turn_on_device(struct pci_dev *pdev, int device);
void vrc4173_turn_off_device(struct pci_dev *pdev, int device);

// I don't know about other devices, but WorkPad doesn't have LEDs for
// the lock states, rather those are handled on screen.  Maybe we should
// send an event notification to user space when LED state changes?  MFK

void kbd_leds(unsigned char leds)
{
	return;
}

// scancode -> keycode translation not supported, since all scancodes
// can be addressed in the keymap

int kbd_setkeycode(unsigned int scancode, unsigned int keycode)
{
	return (scancode == keycode) ? 0 : -EINVAL;
}

int kbd_getkeycode(unsigned int scancode)
{
	return scancode;
}

int kbd_translate(unsigned char scancode, unsigned char *keycode,
	char raw_mode)
{
	*keycode = scancode;
	return 1;
}

char kbd_unexpected_up(unsigned char keycode)
{
	printk(KERN_WARNING "VRC4173 KIU: unexpected up, keycode 0x%x\n", keycode);
	return 0x80;
}

#define DEFAULT_REP_WAIT (HZ/4)
#define DEFAULT_REP_RATE (HZ/25)

static int rep_scancode = -1;
static unsigned long rep_wait = DEFAULT_REP_WAIT;
static unsigned long rep_rate = DEFAULT_REP_RATE;
static unsigned long next_reptime;

inline static void handle_cmp_kiudat_bit(unsigned short kiudat, unsigned short cmp_kiudat,
	int scancode, unsigned short bitmask)
{
	int down;

	// did this bit change?
	if ( cmp_kiudat & bitmask ) {
		down = kiudat & bitmask ? 1 : 0;
		if (down) {
			// on press, set up repeat
			rep_scancode = scancode;
			next_reptime = jiffies + rep_wait;
		} else {
			// on release, stop repeating
			if (rep_scancode == scancode)
				rep_scancode = -1;
		}

		// send the scancode and down or up to higher level code
		printk("keyboard 0x%04x\n", scancode);
		handle_scancode(scancode, down);
	}
}

inline static void handle_cmp_kiudat_nbl(unsigned short kiudat, unsigned short cmp_kiudat,
	int scancode, unsigned short nblmask)
{
	unsigned short bitmask;

	// common case if for nothing set
	if (cmp_kiudat & nblmask) {
		// handle each bit
		for (bitmask = nblmask & 0x1111 ; bitmask & nblmask; bitmask <<= 1)
			handle_cmp_kiudat_bit(kiudat, cmp_kiudat, scancode++, bitmask);
	}
}

inline static void handle_kiudat(unsigned short kiudat, unsigned short last_kiudat,
	int scancode_base)
{
	unsigned short cmp_kiudat, nblmask;
	int scancode = scancode_base;

	// compare this data with the previous data, setting only bits that changed
	cmp_kiudat = kiudat ^ last_kiudat;

	// nothing changed - most common case
	if ( cmp_kiudat == 0 )
		return;

	for (nblmask = 0x000f; nblmask; nblmask <<= 4) {
		// handle each nibble
		handle_cmp_kiudat_nbl(kiudat, cmp_kiudat, scancode, nblmask);
		scancode += 4;
	}
}

#define KIUDAT_REGISTERS 6

inline static void handle_kiu_data_ready(struct pci_dev *pdev)
{
	static unsigned short last_kiudat[KIUDAT_REGISTERS];
	unsigned short thisdat, lastdat;
	int i;
	long ioaddr;

	ioaddr = pci_resource_start (pdev, 0);

	for (i = 0; i < KIUDAT_REGISTERS; i++) {
		lastdat = last_kiudat[i];
		thisdat = last_kiudat[i] = inw(ioaddr + VRC4173_KIUDAT0 + i * sizeof(short));
		handle_kiudat(thisdat, lastdat, i * 16);
	}
	if ((rep_scancode >= 0) && (time_after_eq(jiffies, next_reptime))) {
		next_reptime = jiffies + rep_rate;
		printk("keyboard 0x%04x\n", rep_scancode);
		handle_scancode(rep_scancode, 1);
	}
}

void vrc4173_handle_keyboard(struct pci_dev *pdev, int index)
{
	unsigned short kiuint;
	long ioaddr;

	ioaddr = pci_resource_start (pdev, 0);

	// get keyboard interrupt status
	kiuint = inw(ioaddr + VRC4173_KIUINT);
	barrier();

	// ack them, but only the ones that were read.  if we just ack
	// all of them, we run the risk of ignoring interrupts that
	// happen between these 2 lines
	outw(kiuint, ioaddr + VRC4173_KIUINT);

	// keyboard data ready?
	if ( kiuint & VRC4173_KIUINT_KDATRDY ) {
	    handle_kiu_data_ready(pdev);
	}

	// lost keyboard data?
	if ( kiuint & VRC4173_KIUINT_KDATLOST ) {
		printk("VRC4173 KIU: keyboard data lost\n");
		printk(KERN_DEBUG "VRC4173 KIU: keyboard data lost\n");
	}
}

int vrc4173_keyboard_init(struct pci_dev *pdev, int channel)
{
	long ioaddr;

	ioaddr = pci_resource_start (vrc4173_dev, 0);

	printk("VRC4173 KIU keyboard driver version 0.1\n");

	vrc4173_turn_on_device(vrc4173_dev, KIU_DEV);

	// use GPIO[0-15] for keyboard scanning
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_SELECTREG, 0xfffe, 0);
	
	// reset the KIU
	outw(0x0001, ioaddr + VRC4173_KIURST);

	barrier();

	// number of scanlines to use varies by device, but most use all 12
#if (KBD_SCANLINES == 8)
	outw(0x0002, ioaddr + VRC4173_SCANLINE);
#else
#error "Illegal number of scanlines defined in KBD_SCANLINES"
#endif

	// enable scan, auto scan, auto stop, and one stop rep
	outw(VRC4173_KIUSCANREP_KEYEN | VRC4173_KIUSCANREP_ATSCAN |
	     VRC4173_KIUSCANREP_ATSTP | (1 << 4),
	     ioaddr + VRC4173_KIUSCANREP);

	// set stabilization, off, and wait times in 30us increments
	// this uses a short (but not the shortest) setting for each
	outw((9-1) | ((9-1) << 5) | ((9-1) << 10), ioaddr + VRC4173_KIUWKS);

	// set wait time between scan cycles in 30us increments
	// setting this too short can result in lost keyboard data
	// this wait is on top of the stabilization, off, and wait times above,
	// which are per scan line, so total cycle time is about 1/50 sec
	outw((20000 / 30) - KBD_SCANLINES * 27, ioaddr + VRC4173_KIUWKI);

	barrier();

	// clear any pending interrupts
	outw(0x0007, ioaddr + VRC4173_KIUINT);

	barrier();

	// unmask the keyboard interrupts that we are interested in
	outw(VRC4173_KIUINT_KDATRDY | VRC4173_KIUINT_KDATLOST, ioaddr + VRC4173_MKIUINTREG);

	return 0;
}

void __init kbd_init_hw(void)
{
    // Wait for the PCI system to be initialized and the VRC4173 found that way
}
