/*
 *  linux/drivers/ide/necide.c -- NEC uPD61130 (EMMA2) IDE Driver
 *
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/types.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/delay.h>
#include <linux/ide.h>
#include <linux/init.h>
#include <mmac/hardware/hwata.h>
#include <mmac/hardware/hwxdma.h>
#include <asm/io.h>

int nec_ide_dmaproc(ide_dma_action_t func, ide_drive_t *drive);
static void DMA_SetRegister( unsigned int* dmaRegister, unsigned int value );
    /*
     *  Base of the IDE interface (see ATAManager ROM code)
     */

#define NEC_HD_BASE	0x2000

    /*
     *  Offsets from the above base (scaling 4)
     */

// offsets for 16-bit access
#define NEC_ATA_CONTROL 0x2
#define NEC_ATA_INT_EN_ITOP_STATUS 0x62
#define NEC_ATA_INT_STATUS 0x66
#define NEC_ATA_MODE 0x6a

// standard ata registers
#define NEC_HD_DATA	0xd22
#define NEC_HD_ERROR	0xd26		/* see err-bits */
#define NEC_HD_NSECTOR	0xd2a		/* nr of sectors to read/write */
#define NEC_HD_SECTOR	0xd2e		/* starting sector */
#define NEC_HD_LCYL	0xd32		/* starting cylinder */
#define NEC_HD_HCYL	0xd36		/* high byte of starting cyl */
#define NEC_HD_SELECT	0xd3a		/* 101dhhhh , d=drive, hhhh=head */
#define NEC_HD_STATUS	0xd3e		/* see status-bits */
#define NEC_HD_CONTROL	0xd5a		/* control/altstatus */

static int __init necide_offsets[IDE_NR_PORTS] = {
    NEC_HD_DATA, NEC_HD_ERROR, NEC_HD_NSECTOR, NEC_HD_SECTOR, NEC_HD_LCYL,
    NEC_HD_HCYL, NEC_HD_SELECT, NEC_HD_STATUS, NEC_HD_CONTROL
};

#define NEC_HD_ISR	0x66


/* #defines for List DMA Reset */
#define EMMA_BIOS_DMA_LIST_DMA_RST_MEM_BIT_MASK		(0x00000040)
#define EMMA_BIOS_DMA_LIST_DMA_RST_GIOW_BIT_MASK	(0x00000020)
#define EMMA_BIOS_DMA_LIST_DMA_RST_GIOR_BIT_MASK	(0x00000010)
#define EMMA_BIOS_DMA_LIST_DMA_RST_IDEW_BIT_MASK	(0x00000008)
#define EMMA_BIOS_DMA_LIST_DMA_RST_IDER_BIT_MASK	(0x00000004)
#define EMMA_BIOS_DMA_LIST_DMA_RST_PCMW_BIT_MASK	(0x00000002)
#define EMMA_BIOS_DMA_LIST_DMA_RST_PCMR_BIT_MASK	(0x00000001)
#define EMMA_BIOS_DMA_LIST_DMA_RSTALL_BIT_MASK		(0x0000007F)

/* #defines for PRI_CHSEL channel priorities - DMA channel IDs */
#define EMMA_BIOS_DMA_VIDMAIN_CHAN_ID	(0x00000000)
#define EMMA_BIOS_DMA_VIDSUB_CHAN_ID	(0x20000000)
#define EMMA_BIOS_DMA_AUDCH0_CHAN_ID	(0x40000000)
#define EMMA_BIOS_DMA_AUDCH1_CHAN_ID	(0x60000000)
#define EMMA_BIOS_DMA_PCMR_CHAN_ID		(0x80000000)
#define EMMA_BIOS_DMA_PCMW_CHAN_ID		(0x90000000)
#define EMMA_BIOS_DMA_IDER_CHAN_ID		(0xA0000000)
#define EMMA_BIOS_DMA_IDEW_CHAN_ID		(0xB0000000)
#define EMMA_BIOS_DMA_MEM_CHAN_ID		(0xC0000000)
#define EMMA_BIOS_DMA_GIORW_CHAN_ID		(0xE0000000)

/* #defines for REQ_ENA_LIST bits */
#define EMMA_BIOS_DMA_REQ_ENA_MEM_BIT_MASK	(0x00000020)
#define EMMA_BIOS_DMA_REQ_ENA_GIO_BIT_MASK	(0x00000010)
#define EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK	(0x00000008)
#define EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK	(0x00000004)
#define EMMA_BIOS_DMA_REQ_ENA_PCMW_BIT_MASK	(0x00000002)
#define EMMA_BIOS_DMA_REQ_ENA_PCMR_BIT_MASK	(0x00000001)

/* #defines for REQ_MASK bits */
#define EMMA_BIOS_DMA_REQ_MSK_MEM_BIT_MASK			(0x00004000)
#define EMMA_BIOS_DMA_REQ_MSK_GIOW_BIT_MASK			(0x00002000)
#define EMMA_BIOS_DMA_REQ_MSK_GIOR_BIT_MASK			(0x00001000)
#define EMMA_BIOS_DMA_REQ_MSK_IDEW_BIT_MASK			(0x00000800)
#define EMMA_BIOS_DMA_REQ_MSK_IDER_BIT_MASK			(0x00000400)
#define EMMA_BIOS_DMA_REQ_MSK_PCMW_BIT_MASK			(0x00000200)
#define EMMA_BIOS_DMA_REQ_MSK_PCMR_BIT_MASK			(0x00000100)
#define EMMA_BIOS_DMA_LIST_DMA_ENAREQALL_BIT_MASK	(0x00007F0F)

/* #define for LIST DMA disable bit */
#define EMMA_BIOS_DMA_NOLIST_BIT_MASK				(0x00000001)
													
/* #defines for non-List DMA start and done bits */
#define EMMA_BIOS_DMA_DMA_START_BIT_MASK			(0x00000001)
#define EMMA_BIOS_DMA_DMA_DONE_BIT_MASK				(0x00000002)

/* #defines for INTERRUPT bits */
#define EMMA_BIOS_DMA_INT_LIST_ERROR_BIT_MASK	(0x04000000)
#define EMMA_BIOS_DMA_INT_REQ_ERROR_BIT_MASK	(0x02000000)
#define EMMA_BIOS_DMA_INT_MEM_DONE_BIT_MASK		(0x00400000)
#define EMMA_BIOS_DMA_INT_GIOW_DONE_BIT_MASK	(0x00200000)
#define EMMA_BIOS_DMA_INT_GIOR_DONE_BIT_MASK	(0x00100000)
#define EMMA_BIOS_DMA_INT_IDEW_DONE_BIT_MASK	(0x00080000)
#define EMMA_BIOS_DMA_INT_IDER_DONE_BIT_MASK	(0x00040000)
#define EMMA_BIOS_DMA_INT_PCMW_DONE_BIT_MASK	(0x00020000)
#define EMMA_BIOS_DMA_INT_PCMR_DONE_BIT_MASK	(0x00010000)

/* Define the number of DMA Channels */
#define NUM_DMA_CHANNELS                    (12)

/* Define the max value of DMA Channel Priority  */
/* Note - priority is from 0-9, with 0 being the */
/*        highest and 9 the lowest.              */
#define MAX_DMA_CHANNEL_PRIORITY			(9)
/* Define the number of bits used in a DMA channel ID */
/* (used for setting the DMA channel priorities)      */
#define BITS_IN_DMA_CHANNEL_ID				(4)
/* Define the number of bits used to set the wait */
/* limit for each DMA channel priority            */
#define BITS_IN_DMA_CHANNEL_WAIT_LIMIT		(4)

#define MIN_DMA_CHANNEL_PRIORITY_BIT_MASK 	(0xF0000000)
#define DMA_CHANNEL_PRIORITY_8_BIT_MASK 	(0xF0000000)
#define MAX_DMA_CHANNEL_PRIORITY_BIT_MASK 	(0x0F000000)

/* Define a value to be used to set a burst length of 1 */
#define MMAC_DMA_BURST_LENGTH_ONE           (1)

/* Define used to enable/disable use of direct register
   DMA (no-list DMA) by the function MMAC_DMA_StartTransfer */
#define USE_NO_LIST_DMA						(MMAC_YES)

/* Define used to enable/disable Memory-To-Memory DMA
   - if disabled any attempt to perform Mem-To-Mem DMA by
   calling either MMAC_DMA_StartTransfer() or 
   MMAC_DMA_AddListDMAList() will result in the value
   MMAC_DMA_INVALID_PARAM being returned.				    */
#define ALLOW_MEM_TO_MEM_DMA				(MMAC_YES)

#define MMAC_ATA_USE_GPIO_PIN_FOR_ISR       (MMAC_NO)
/* The #define below informs the ATA driver which
   GPIO pin will be used to trigger an ATA interrupt 
   (e.g. 9 = PPORT9).   
*/
#define MMAC_ATA_GPIO_PIN_FOR_ISR           (9)
/* The #define below informs the ATA driver when the
   specified GPIO pin will trigger an ATA interrupt 
   (e.g. rising edge, both edges, high level, etc).   
   The value must be a defined value for the type
   MMAC_GPIO_INTERRUPT_MODE.
*/
#define MMAC_ATA_GPIO_PIN_MODE              (MMAC_GPIO_RISING_EDGE)
 
/* #defines for ATA COMMAND CODES */
#define MMAC_ATA_ATA_ATA_IDENTIFY_DEVICE	(0x000000EC)
#define MMAC_ATA_ATA_ATA_READ_DMA			(0x000000C8)
#define MMAC_ATA_ATA_ATA_WRITE_DMA			(0x000000CA)
#define MMAC_ATA_ATA_ATA_WRITE_SECTORS		(0x00000030)
#define MMAC_ATA_ATA_ATA_READ_SECTORS		(0x00000020)
#define MMAC_ATA_ATA_ATA_READ_MULTIPLE		(0x000000C4)
#define MMAC_ATA_ATA_ATA_WRITE_MULTIPLE	    (0x000000C5)
#define MMAC_ATA_ATA_SEEK					(0x00000070)
#define	MMAC_ATA_ATA_SET_FEATURES			(0x000000EF)

#define MMAC_ATA_ATA_SET_TRANSFER_MODE		(0x00000003)
#define MMAC_ATA_ATA_UDMA_MODE				(0x00000040)
#define MMAC_ATA_ATA_DMA_MODE				(0x00000020)
#define MMAC_ATA_ATA_PIO_MODE_IORDY		    (0x00000000)
#define MMAC_ATA_ATA_PIO_MODE_NO_IORDY		(0x00000001)

/* #defines for ATA_CONTROL register */
#define MMAC_ATA_ATA_UDMA_TRANSFER_BIT_MASK	(0x00000006)
#define MMAC_ATA_ATA_DIR_BIT_MASK			(0x00000008)

/* #defines for DMA/UDMA Timing Modes */
#define MMAC_ATA_ATA_CFG_DMA_MODE_0		    (0x00003333)/*(0x00007777)*/
#define MMAC_ATA_ATA_CFG_DMA_MODE_1		    (0x00003333)/*(0x00002222)*/
#define MMAC_ATA_ATA_CFG_DMA_MODE_2		    (0x00001111)/*(0x00002020)*/
#define MMAC_ATA_ATA_CFG_DMA_DEFAULT		(0x00000000)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_0		(0x00000333)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_1		(0x00000322)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_2		(0x00000311)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_4		(0x00000311)
#define MMAC_ATA_ATA_CFG_UDMA_DEFAULT		(0x00000077)/*(0x00000377)*/

/* #defines for PIO control register */
#define MMAC_ATA_ATA_S_POE_BIT_MASK		    (0x00000010)
#define MMAC_ATA_ATA_M_POE_BIT_MASK		    (0x00000020)
#define MMAC_ATA_ATA_S_PRE_BIT_MASK		    (0x00000040)
#define MMAC_ATA_ATA_M_PRE_BIT_MASK		    (0x00000080)
#define MMAC_ATA_ATA_MAX_PIO_CNT_MASK		(0x0000FF00)

/* #defines for PIO Mode register */
#define MMAC_ATA_ATA_PIO_MASK				(0x0000000F)
#define MMAC_ATA_ATA_RDY_EN_BIT_MASK		(0x00008000)

/* #defines for DMA/UDMA registers */
#define MMAC_ATA_ATA_DMA_M_CMD_MASK		    (0x0000F000)
#define MMAC_ATA_ATA_DMA_M_RCV_MASK		    (0x00000F00)
#define MMAC_ATA_ATA_DMA_S_CMD_MASK		    (0x000000F0)
#define MMAC_ATA_ATA_DMA_S_RCV_MASK		    (0x0000000F)
#define MMAC_ATA_ATA_MUDM_BIT_MASK			(0x00000100)
#define MMAC_ATA_ATA_SUDM_BIT_MASK			(0x00000200)
#define MMAC_ATA_ATA_UDMA_M_COM_MASK		(0x00000070)
#define MMAC_ATA_ATA_UDMA_S_COM_MASK		(0x00000007)
#define MMAC_ATA_ATA_DMA_ON_BIT_MASK		(0x00000001)
#define MMAC_ATA_UDMA_66_ON_BIT_MASK		(0x00000100)

/* #defines for LBA register(s) */
#define MMAC_ATA_ATA_BITS_7TO0_MASK		    (0x000000FF)
#define MMAC_ATA_ATA_BITS_15TO8_MASK		(0x0000FF00)
#define MMAC_ATA_ATA_BITS_23TO16_MASK		(0x00FF0000)
#define MMAC_ATA_ATA_BITS_27TO24_MASK		(0x0F000000)
#define MMAC_ATA_ATA_LBA_27TO24_MASK		(0x0000000F)
#define MMAC_ATA_ATA_DRV_BIT_MASK			(0x00000010)
#define MMAC_ATA_ATA_LBA_BIT_MASK			(0x00000040)

/* #defines for Bus Master register */
#define MMAC_ATA_ATA_BM_EN_BIT_MASK		    (0x00000040)
#define MMAC_ATA_ATA_BM_IO_AI_BIT_MASK		(0x00000020)
#define MMAC_ATA_ATA_BM_MEMR_MODE_MASK		(0x00000018)
#define MMAC_ATA_ATA_BM_MIO_BIT_MASK		(0x00000004)
#define MMAC_ATA_ATA_DISC_EN_BIT_MASK		(0x00000002)
#define MMAC_ATA_ATA_BM_SSC_BIT_MASK		(0x00000001)

/* #defines for byte/word order swap register */
#define MMAC_ATA_ATA_BYTE_SWAP_PB_BIT_MASK	(0x00000200)
#define MMAC_ATA_ATA_WORD_SWAP_PB_BIT_MASK	(0x00000100)
#define MMAC_ATA_ATA_BYTE_SWAP_REC_BIT_MASK	(0x00000002)
#define MMAC_ATA_ATA_WORD_SWAP_REC_BIT_MASK	(0x00000001)

/* #define for ATA Soft Reset bit */
#define MMAC_ATA_ATA_ATA_SOFT_RESET_SRST_BIT_MASK	(0x00000001)

/* #define for ATA interrupt bit */
#define MMAC_ATA_ATA_ATA_INT_STATUS_INT_BIT_MASK	(0x00000001)
#define MMAC_ATA_ATA_ATA_INT_ENABLE_BIT_MASK		(0x00004000)

/* #define for ATA IDE Status/ Alt Status error bit */
#define MMAC_ATA_ATA_ATA_IDE_STATUS_ERR_BIT_MASK	(0x00000001)

/* #define for ATA IDE Features/ Error ABRT bit */
#define MMAC_ATA_ATA_ATA_IDE_FEATURES_ERROR_ABRT_BIT_MASK	(0x00000004)

/* #defines for driveInfo words */
#define MMAC_ATA_ATA_DRIVE_INFO_PIO_MODE_BIT_MASK 			(0xFF00)
#define MMAC_ATA_ATA_DRIVE_INFO_IORDY_SUPPORTED_BIT_MASK	(0x0800)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_2_BIT_MASK		    (0x0004)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_1_BIT_MASK		    (0x0002)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_0_BIT_MASK		    (0x0001)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_4_BIT_MASK		(0x0010)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_2_BIT_MASK		(0x0004)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_1_BIT_MASK		(0x0002)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_0_BIT_MASK		(0x0001)
#define	MMAC_ATA_ATA_DRIVE_INFO_ADV_PIO_MODE_4_MASK		    (0x0002)
#define	MMAC_ATA_ATA_DRIVE_INFO_ADV_PIO_MODE_3_MASK		    (0x0001)

/* #defines for BHIF ICLK and I2CLK */
#define MMAC_ATA_ICLK_SEL			(0x01000000)
#define MMAC_ATA_I2CLK_SEL			(0x02000000)
#define MMAC_ATA_ICLK_MSK			(0x00002000)
#define MMAC_ATA_I2CLK_MSK			(0x00004000)

/* #define for the number of ATA devices that can be attached (2) */
#define MMAC_ATA_MAX_DEVICES        (2)

/* #define for max. wait period when waiting for NotBsyNotDrq, etc */
#define ATA_WAIT_TIMEOUT            (200000)

#define ATA_ACT_BIT_TIMEOUT         (0xfffff)

/* #Define to enable support for UDMA Mode 4 - if enabled and device
   supports UDMA Mode 4, this will be used automatically.
   If disabled, UDMA Mode 2 and below only will be supported by the
   ATA driver. */
/* (Note - due to the clock settings permitted on EMMA2, it is not
    possible for us to support UDMA Mode 3 ). */
#define ATA_SUPPORT_UDMA_MODE_4		(MMAC_NO)

/* #Define for the bits set in the ATA_LEN register following
   a DMA tranfser, which indicate that a pause condition occurred
   which will have caused data to be duplicated when reading.
   This is indicated in bits 20, 19 and 18. */
#define ATA_LEN_REGISTER_ERR_BITS	(0x001C0000)

static unsigned int nec_dma_addr[128];
static unsigned int nec_dma_len[128];
static unsigned int nec_dma_count=0;

static int nec_ack_intr(ide_hwif_t* hwif)
{
	unsigned char isr;
	isr = inw(NEC_HD_BASE + NEC_HD_ISR);
	if (isr & 1) {
		outw(isr & ~1, NEC_HD_BASE + NEC_HD_ISR);
		return 1;
	}

	return 0;
}
#define NEC_HD_IRQ 13

void /*__init */ necide_init(void)
{
	hw_regs_t hw;
	int index = -1;


	// check machine type
	// return if not correct
	printk("nec_ide_init\n");

    ide_setup_ports(&hw, (ide_ioreg_t)NEC_HD_BASE, necide_offsets,
		    0, 0, NULL, NEC_HD_IRQ);
    index = ide_register_hw(&hw, NULL);
	printk("nec_ide_init2\n");
    /*
    ide_hwifs[index].tuneproc = pmac_ide_tuneproc;
    ide_hwifs[index].selectproc = pmac_ide_selectproc;
    ide_hwifs[index].speedproc = &pmac_ide_tune_chipset;
    */
    ide_hwifs[index].dmaproc = &nec_ide_dmaproc;
#ifdef CONFIG_BLK_DEV_IDEDMA_PMAC_AUTO
		if (!noautodma)
			ide_hwifs[ix].autodma = 1;
#endif
	printk("nec_ide_init done\n");

}

/*
 * pmac_ide_build_dmatable builds the DBDMA command list
 * for a transfer and sets the DBDMA channel to point to it.
 */
static int
nec_ide_build_dmatable(ide_drive_t *drive, int ix, int wr)
{
	int count = 0;
	struct request *rq = HWGROUP(drive)->rq;
	struct buffer_head *bh = rq->bh;
	unsigned int size, addr;
	nec_dma_count=0;

	do {
		/*
		 * Determine addr and size of next buffer area.  We assume that
		 * individual virtual buffers are always composed linearly in
		 * physical memory.  For example, we assume that any 8kB buffer
		 * is always composed of two adjacent physical 4kB pages rather
		 * than two possibly non-adjacent physical 4kB pages.
		 */
		if (bh == NULL) {  /* paging requests have (rq->bh == NULL) */
			addr = virt_to_bus(rq->buffer);
			size = rq->nr_sectors << 9;
		} else {
			/* group sequential buffers into one large buffer */
			addr = virt_to_bus(bh->b_data);
			size = bh->b_size;
			while ((bh = bh->b_reqnext) != NULL) {
				if ((addr + size) != virt_to_bus(bh->b_data))
					break;
				size += bh->b_size;
			}
		}

		/*
		 * Fill in the next DBDMA command block.
		 * Note that one DBDMA command can transfer
		 * at most 65535 bytes.
		 */
		/* use addr , size */
		nec_dma_addr[nec_dma_count]=addr;
		nec_dma_len[nec_dma_count++]=size;

		   printk("dma: addr=%x size=%x\n",addr,size);
	} while (bh != NULL);

	return 1;
}

/*
 * dma_intr() is the handler for disk read/write DMA interrupts
 */
ide_startstop_t nec_ide_dma_intr (ide_drive_t *drive)
{
	int i;
	byte stat, dma_stat;
	printk("ida_dma_intr\n");

	dma_stat = HWIF(drive)->dmaproc(ide_dma_end, drive);
	stat = GET_STAT();			/* get drive status */
	if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
		if (!dma_stat) {
			struct request *rq = HWGROUP(drive)->rq;
			rq = HWGROUP(drive)->rq;
			for (i = rq->nr_sectors; i > 0;) {
				i -= rq->current_nr_sectors;
				ide_end_request(1, HWGROUP(drive));
			}
			return ide_stopped;
		}
		printk("%s: dma_intr: bad DMA status (dma_stat=%x)\n", 
		       drive->name, dma_stat);
	}
	return ide_error(drive, "dma_intr", stat);
}

static int
nec_ide_check_dma(ide_drive_t *drive)
{
	int ata4, udma, idx;
	struct hd_driveid *id = drive->id;
	int enable = 1;

	drive->using_dma = 0;
	
	if (drive->media == ide_floppy)
		enable = 0;
	if (((id->capability & 1) == 0) )
		enable = 0;

	udma = 0;
	// ata4 = (pmac_ide[idx].kind == controller_kl_ata4);
	if(enable) {
	    printk("set using_dma\n");
	    drive->using_dma = 1;
#if 0		
		if (ata4 && (drive->media == ide_disk) &&
		    (id->field_valid & 0x0004) && (id->dma_ultra & 0x17)) {
			/* UltraDMA modes. */
			drive->using_dma = pmac_ide_udma_enable(drive, idx);
		}
		if (!drive->using_dma && (id->dma_mword & 0x0007)) {
			/* Normal MultiWord DMA modes. */
			drive->using_dma = pmac_ide_mdma_enable(drive, idx);
		}
		/* Without this, strange things will happen on Keylargo-based
		 * machines
		 */
		OUT_BYTE(0, IDE_CONTROL_REG);
		/* Apply settings to controller */
		pmac_ide_selectproc(drive);
#endif
	}
	return 0;
}


static int nec_dma_start_transfer(unsigned int addr,int length,int do_write)
{
    unsigned int tempValue;
    printk("nec_start_dma %x %x %d\n",addr,length,do_write);
    emmaATA.ATA_LEN = length;
    if(do_write) {
	emmaATA.ATA_CONTROL = 0/* read */ |  3;
					/* Reset ATA-to-MEM transfers... */
	                tempValue = emmaXDMA.XDMA_LIST_RST;
	                tempValue |= EMMA_BIOS_DMA_LIST_DMA_RST_IDEW_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_LIST_RST, tempValue );

					/* Dummy read to implement short delay */
	    		tempValue = emmaXDMA.XDMA_REQ_STAT_LIST;

	                tempValue = emmaXDMA.XDMA_LIST_RST;
	                tempValue &= ~EMMA_BIOS_DMA_LIST_DMA_RST_IDEW_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_LIST_RST, tempValue );

	                /* Set destination end address */
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_W_BUF_END_ADR,
                    				 (((UI32)(addr+length-1) & 0x03FFFFFF) >> 3) );

	                /* Set destination start address */
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_W_BUF_STA_ADR,
                    				 (((UI32)addr & 0x03FFFFFF) >> 3) );
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_W_PTR,
                    				 (((UI32)addr & 0x03FFFFFF) >> 3) );

					/* Set IDEW_COMMAND register last */
					tempValue = ((UI32)MMAC_DMA_DEFAULT_BURST_LENGTH & 0x0000000F) << 28;
					/* set inten and nolist bits */
					tempValue |= 0x02000001;
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_W_COMMAND, tempValue );

					/* Enable ATA-to-MEM DMA requests */
	                tempValue = emmaXDMA.XDMA_REQ_ENA_LIST;
	                tempValue |= EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_REQ_ENA_LIST, tempValue );
					/* Enable ATA-to-MEM DMA requests */
	                tempValue = emmaXDMA.XDMA_REQ_MASK;
	                tempValue |= 0x00000800; /* Allow requests */
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_REQ_MASK, tempValue );
     } else { 
	emmaATA.ATA_CONTROL = 8 /* read */ |  3;
					/* Reset MEM-to-ATA transfers... */
	                tempValue = emmaXDMA.XDMA_LIST_RST;
	                tempValue |= EMMA_BIOS_DMA_LIST_DMA_RST_IDER_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_LIST_RST, tempValue );

					/* Dummy read to implement short delay */
	    		    tempValue = emmaXDMA.XDMA_REQ_STAT_LIST;

	                tempValue = emmaXDMA.XDMA_LIST_RST;
	                tempValue &= ~EMMA_BIOS_DMA_LIST_DMA_RST_IDER_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_LIST_RST, tempValue );

	                /* Set source end address */
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_R_BUF_END_ADR,
                    				 (((UI32)(addr+length-1) & 0x03FFFFFF) >> 3) );

	                /* Set source start address */
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_R_BUF_STA_ADR,
                    				 (((UI32)addr & 0x03FFFFFF) >> 3) );
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_R_PTR,
                    				 (((UI32)addr & 0x03FFFFFF) >> 3) );

					/* Set IDER_COMMAND register last */
					tempValue = ((UI32)MMAC_DMA_DEFAULT_BURST_LENGTH & 0x0000000F) << 28;
					/* set inten and nolist bits */
					tempValue |= 0x02000001;
                    DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_IDE_R_COMMAND, tempValue );

					/* Enable MEM-to-ATA DMA requests */
	                tempValue = emmaXDMA.XDMA_REQ_ENA_LIST;
	                tempValue |= EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK;
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_REQ_ENA_LIST, tempValue );
					/* Enable MEM-to-ATA DMA requests */
	                tempValue = emmaXDMA.XDMA_REQ_MASK;
	                tempValue |= 0x00000400; /* Allow requests */
	                DMA_SetRegister( (UI32*)&emmaXDMA.XDMA_REQ_MASK, tempValue );
				}
}

int nec_ide_dmaproc(ide_dma_action_t func, ide_drive_t *drive)
{
	int ix, dstat, i;
	volatile struct dbdma_regs *dma;

	/* Can we stuff a pointer to our intf structure in config_data
	 * or select_data in hwif ?
	 */

	switch (func) {
	case ide_dma_off:
		printk(KERN_INFO "%s: DMA disabled\n", drive->name);
	case ide_dma_off_quietly:
		drive->using_dma = 0;
		break;
	case ide_dma_on:
	case ide_dma_check:
		nec_ide_check_dma(drive);
		break;
	case ide_dma_read:
		printk("ide_dma_read\n");
	case ide_dma_write:
		if(func == ide_dma_write) printk("ide_dma_write\n");
		if (!nec_ide_build_dmatable(drive, ix, func==ide_dma_write))
			return 1;
		if(nec_dma_count>1) {
		    printk("dma: currently only one transfer\n");
		    return 0;
		}
		drive->waiting_for_dma = 1;
		nec_dma_start_transfer(nec_dma_addr[0],nec_dma_len[0],func==ide_dma_write);
	        emmaATA.ATA_DMA_START = MMAC_ATA_ATA_DMA_ON_BIT_MASK; 
		if (drive->media != ide_disk)
			return 0;
		ide_set_handler(drive, &nec_ide_dma_intr, WAIT_CMD, NULL);
		OUT_BYTE(func==ide_dma_write? WIN_WRITEDMA: WIN_READDMA,
			 IDE_COMMAND_REG);
		if(func==ide_dma_write)
		    emmaXDMA.XDMA_REQ_MASK |= EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK;
		else
		    emmaXDMA.XDMA_REQ_MASK |= EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK;
	case ide_dma_begin:
		printk("ide_dma_begin\n");
		break;
	case ide_dma_end:
		printk("ide_dma_end\n");
		emmaXDMA.XDMA_REQ_MASK &= ~EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK;
		emmaXDMA.XDMA_REQ_MASK &= ~EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK;
		drive->waiting_for_dma = 0;
		return 1; // return 1 for done
	case ide_dma_test_irq:
#if 0
		if ((in_le32(&dma->status) & (RUN|ACTIVE)) == RUN)
			return 1;
		/* That's a bit ugly and dangerous, but works in our case
		 * to workaround a problem with the channel status staying
		 * active if the drive returns an error
		 */
		if (IDE_CONTROL_REG) {
			byte stat;
			stat = GET_ALTSTAT();
			if (stat & ERR_STAT)
				return 1;
		}
		/* In some edge cases, some datas may still be in the dbdma
		 * engine fifo, we wait a bit for dbdma to complete
		 */
		while ((in_le32(&dma->status) & (RUN|ACTIVE)) != RUN) {
			if (++i > 100)
				return 0;
			udelay(1);
		}
#endif
		return 1;
#if 0
		/* Let's implement tose just in case someone wants them */
	case ide_dma_bad_drive:
	case ide_dma_good_drive:
		return check_drive_lists(drive, (func == ide_dma_good_drive));
	case ide_dma_verbose:
		return report_drive_dmaing(drive);
#endif
	case ide_dma_retune:
	case ide_dma_lostirq:
	case ide_dma_timeout:
		printk(KERN_WARNING "ide_nec_dmaproc: chipset supported %s func only: %d\n", ide_dmafunc_verbose(func),  func);
		return 1;
	default:
		printk(KERN_WARNING "ide_nec_dmaproc: unsupported %s func: %d\n", ide_dmafunc_verbose(func), func);
		return 1;
	}
	return 0;
}

static void DMA_SetRegister( unsigned int* dmaRegister, unsigned int value )
{
    *dmaRegister = value;
    
    while( *dmaRegister != value )
    {
        *dmaRegister = value;
    }
}
