/*
 * LIRC base driver
 * 
 * (L) by Artur Lipowski <alipowski@kki.net.pl>
 *        This code is licensed under GNU GPL
 *
 * $Id: lirc_dev.c,v 1.17 2000/12/23 11:23:57 columbus Exp $
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
 
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define LIRC_HAVE_DEVFS
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 4)
#error "********************************************************"
#error " Sorry, this driver needs kernel version 2.2.4 or higher"
#error "********************************************************"
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/poll.h>
#ifdef LIRC_HAVE_DEVFS
#include <linux/devfs_fs_kernel.h>
#endif
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/errno.h>
#include <linux/wrapper.h>
#include <mmac/types.h>
#include <mmac/capcom.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "drivers/lirc.h"

#include "lirc_dev.h"

static int debug = 0;

MODULE_PARM(debug,"i");

#define IRCTL_DEV_NAME    "EmmaLircDev"
#define IRLOCK            down_interruptible(&ir->lock)
#define IRUNLOCK          up(&ir->lock)
#define SUCCESS           0
#define NOPLUG            -1
#define dprintk           if (debug) printk

#define LOGHEAD           "lirc_dev (%s[%d]): "

static struct file_operations fops;
static DECLARE_WAIT_QUEUE_HEAD(lirc_wait_in);

#define RBUF_LEN 256
static lirc_t rbuf[RBUF_LEN];
static int rbh, rbt;

static int capcom_rptr=0;

/* value used to store which of the four transitions reported by CAPCOM driver */
/* we should begin reading from, in case we previously timed-out and manually  */
/* read some of them already... */

/* UI8 used to store the current subsystem status... */
#define IR_CAPCOM_TIMER 0


/* MMAC_BOOL used to store the current signal state... */
static MMAC_BOOL infraRedState = MMAC_FALSE;
static MMAC_CAPCOM_OPEN_IOPB	irInputConfig;


/****************************************************************/
/*			FORWARD DECLARATIONS			*/
/****************************************************************/

/* The Infra Red driver's callback function for the */
/* Capture/Compare driver...                        */
static void IR_CapcomCallback ( MMAC_CAPCOM_CALLBACK_STRUCT *cs );



#define RBWRITE frbwrite

static void inline rbwrite(lirc_t l)
{
	unsigned int nrbt;

	nrbt=(rbt+1) & (RBUF_LEN-1);
	if(nrbt==rbh)      /* no new signals will be accepted */
	{
#               ifdef DEBUG
		printk(KERN_WARNING  LIRC_DRIVER_NAME  ": Buffer overrun\n");
#               endif
		return;
	}
	rbuf[rbt]=l;
	rbt=nrbt;
	wake_up_interruptible(&lirc_wait_in);
}

static void inline frbwrite(lirc_t l)
{
	/* simple noise filter */
	static lirc_t pulse=0L,space=0L;
	static unsigned int ptr=0;
	
	if(ptr>0 && (l&PULSE_BIT))
	{
		pulse+=l&PULSE_MASK;
		if(pulse>50)
		{
			rbwrite(space);
			rbwrite(pulse|PULSE_BIT);
			ptr=0;
			pulse=0;
		}
		return;
	}
	if(!(l&PULSE_BIT))
	{
		if(ptr==0)
		{
			if(l>20000)
			{
				space=l;
				ptr++;
				return;
			}
		}
		else
		{
			if(l>20000)
			{
				space+=pulse;
				if(space>PULSE_MASK) space=PULSE_MASK;
				space+=l;
				if(space>PULSE_MASK) space=PULSE_MASK;
				pulse=0;
				return;
			}
			if(space)
			   rbwrite(space);
			if(pulse)
			   rbwrite(pulse|PULSE_BIT);
			ptr=0;
			pulse=0;
		}
	}
	rbwrite(l);
}

/*****************************************************************
*
* Function Name
*	IR_CapcomCallback
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	
* Description
*	Callback function for input capture/compare timer.
*	This function is called each time there have been four
*	transitions on the IR input.
*	When this occurs, the function calculates the number of
*	bits that have been received (based on the period between
*	transitions), and writes the appropriate number of entries
*	into the FIFO buffer.
*
*****************************************************************/

static void IR_CapcomCallback ( MMAC_CAPCOM_CALLBACK_STRUCT *cs ) 
{
    UI8				transition;
    int i;
    MMAC_CAPCOM_EVENT_BATCH *batch;
    unsigned int period[4];
    batch = &cs->capturedEvents;
    
    
    
    /* Period contains up to 4 transitons - however, some of these may have    */
    /* already been read by calling MMAC_CAPCOM_Read(). This would be the case */
    /* if data finished with two odd transitions which left the CAPCOM driver  */
    /* waiting for two further transitions before notifying us. We then time   */
    /* out, and maually read the two transitions.                              */
    /* When transitions begin again, they fill transitions 3 and 4 in the      */
    /* 'period' field, and then we are notified of 4 transitions (including    */
    /* the two we have already read).                                          */
    /* Therefore, we need to keep track of which transitions of the four we    */
    /* still need to read...                                                   */
    /* Read any previously un-read transitions from 'period'... */

    if(!batch->numEvents) 
	batch->numEvents=4;

    for(i=0;i<batch->numEvents;i++)  
	period[i] = (batch->event[i].period * irInputConfig.params.config.input.resolutionNanoSec)/1000;

    for( transition=capcom_rptr; transition<batch->numEvents; transition++ )
    {
		switch(batch->event[transition].trigger) {
		    case MMAC_CAPCOM_RISING_EDGE:
			infraRedState = MMAC_FALSE;
			break;
		    case MMAC_CAPCOM_FALLING_EDGE:
			infraRedState = MMAC_TRUE;
			break;
		    case MMAC_CAPCOM_TIMER_OVERFLOW:
			batch->event[transition].period = PULSE_MASK;	
			infraRedState = MMAC_TRUE;
			break;
		    default:
			printk("invalid trigger\n");
		}
		if(period[transition] > PULSE_MASK)
		   period[transition] = PULSE_MASK;	
		if(!infraRedState)
		    RBWRITE(PULSE_BIT | period[transition]);
		else
		    RBWRITE(period[transition]);

    }
    capcom_rptr=0;
       
}


static int lirc_open(struct inode *inode, struct file *file)
{
    MMAC_CAPCOM_RESULT_CODE	capComResult;
    MOD_INC_USE_COUNT;
    rbh = 0;
    rbt = 0;
    capcom_rptr=0;
    

    /* Set up the capture/compare timer for IR input*/
    /* Nb. This is used to capture transitions...   */
    irInputConfig.deviceNumber 	= (MMAC_CAPCOM_DEV_NUM)IR_CAPCOM_TIMER;
    irInputConfig.params.mode	= MMAC_CAPCOM_SINGLE_EVENT_INPUT;
    irInputConfig.params.autoRestart = MMAC_TRUE;
    irInputConfig.params.config.input.resolutionNanoSec = 1000;
    irInputConfig.params.config.input.eventTrigger = MMAC_CAPCOM_ANY_EDGE;
    irInputConfig.params.config.input.callback	= (MMAC_CAPCOM_CALLBACK)IR_CapcomCallback;
    

    /* Set initial state... */
    infraRedState = MMAC_FALSE;


    /* Open the capture/compare input timer... */
    capComResult = MMAC_CAPCOM_Open( &irInputConfig );
    if( capComResult == MMAC_CAPCOM_OK )
	{
    	}
	else
	{
	    return -1;
	}
    return 0;
}

/*
 *
 */
static int lirc_close(struct inode *inode, struct file *file)
{
    
    
    MMAC_CAPCOM_Close( IR_CAPCOM_TIMER );
    
    MOD_DEC_USE_COUNT;
    return 0;
}

/*
 *
 */
static unsigned int lirc_poll(struct file *file, poll_table * wait)
{
	poll_wait(file, &lirc_wait_in,wait);
	if (rbh!=rbt) {
		return(POLLIN|POLLRDNORM);
	}
	return(0);
}

/*
 *
 */
static int lirc_ioctl(struct inode *node,struct file *filep,unsigned int cmd,
		      unsigned long arg)
{
        int result;
	unsigned long features=LIRC_CAN_REC_MODE2,mode;
	
	switch(cmd)
	{
	case LIRC_GET_FEATURES:
		result=verify_area(VERIFY_WRITE,(unsigned long*) arg,
				   sizeof(unsigned long));
		if(result) return(result);
		put_user(features,(unsigned long *) arg);
		break;
	case LIRC_GET_SEND_MODE:
		result=verify_area(VERIFY_WRITE,(unsigned long *) arg,
				   sizeof(unsigned long));
		if(result) return(result);
		put_user(LIRC_MODE_PULSE,(unsigned long *) arg);
		break;
	case LIRC_GET_REC_MODE:
		result=verify_area(VERIFY_WRITE,(unsigned long *) arg,
				   sizeof(unsigned long));
		if(result) return(result);
		put_user(LIRC_MODE_MODE2,(unsigned long *) arg);
		break;
	case LIRC_SET_SEND_MODE:
		result=verify_area(VERIFY_READ,(unsigned long *) arg,
				   sizeof(unsigned long));
		if(result) return(result);
		get_user(mode,(unsigned long *) arg);
		if(mode!=LIRC_MODE_PULSE) return(-EINVAL);
		break;
	case LIRC_SET_REC_MODE:
		result=verify_area(VERIFY_READ,(unsigned long *) arg,
				   sizeof(unsigned long));
		if(result) return(result);
		get_user(mode,(unsigned long *) arg);
		if(mode!=LIRC_MODE_MODE2) return(-ENOSYS);
		break;
	default:
		return(-ENOIOCTLCMD);
	}
	return(0);
}

/*
 *
 */

static int lirc_read( struct file *file, char *buf,
		     int count,loff_t *off)
{
	int n=0,retval=0;
	DECLARE_WAITQUEUE(wait,current);
	
	if(n%sizeof(lirc_t)) return(-EINVAL);
	
	add_wait_queue(&lirc_wait_in,&wait);
	current->state=TASK_INTERRUPTIBLE;
	while (n < count)
	{
		if (rbt != rbh) {
			copy_to_user((void *) buf+n,
				     (void *) &rbuf[rbh],sizeof(lirc_t));
			rbh = (rbh + 1) & (RBUF_LEN - 1);
			n+=sizeof(lirc_t);
		} else {
			if (file->f_flags & O_NONBLOCK) {
				retval = -EAGAIN;
				break;
			}
			if(signal_pending(current)) {
			    	retval = -EINTR;
				break;
			}
			schedule();
			current->state=TASK_INTERRUPTIBLE;
		}
	}
	remove_wait_queue(&lirc_wait_in,&wait);
	current->state=TASK_RUNNING;
	return (n ? n : retval);
}

static struct file_operations fops = {
	read:    lirc_read, 
	poll:    lirc_poll,
	ioctl:   lirc_ioctl,
	open:    lirc_open,
	release: lirc_close
};




/*
 *
 */
int lirc_dev_init(void)
{  	
	int i;

#ifndef LIRC_HAVE_DEVFS
 	i = register_chrdev(IRCTL_DEV_MAJOR,
#else
	i = devfs_register_chrdev(IRCTL_DEV_MAJOR,
#endif
				   IRCTL_DEV_NAME,
				   &fops);
	
	if (i < 0) {
		printk ("lirc_dev: device registration failed with %d\n", i);
		return i;
	}
	
	printk("lirc_dev: IR Remote Control driver registered, at major %d \n", 
	       IRCTL_DEV_MAJOR);

	return SUCCESS;
}

/* ---------------------------------------------------------------------- */

/* For now dont try to use it as a static version !  */

#ifdef MODULE

MODULE_DESCRIPTION("LIRC driver module for EMMA2");
MODULE_AUTHOR("Jan Rathert");

/*
 *
 */
int init_module(void)
{
        MMAC_CAPCOM_Initialise();
	return lirc_dev_init();
}

/*
 *
 */
void cleanup_module(void)
{
	int ret;
	
#ifndef LIRC_HAVE_DEVFS
 	ret = unregister_chrdev(IRCTL_DEV_MAJOR, IRCTL_DEV_NAME);
#else
	ret = devfs_unregister_chrdev(IRCTL_DEV_MAJOR, IRCTL_DEV_NAME);
#endif
 
	if (0 > ret){
		printk("lirc_dev: error in module_unregister_chrdev: %d\n",
		       ret);
	} else {
		dprintk("lirc_dev: module successfully unloaded\n");
	}
}
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
