/****************************************************************
*
* FILENAME
*
*	ir.c
*
* PURPOSE 
*
*	EMMA2 Infra Red Driver API
*
* AUTHOR 
*
*	MMAC
*
* DATE 
*
*	02/01/2001
*
* HISTORY 

*   $Date: 9/21/01 3:55 PM $
*
* 	$Revision: \main\9 $

*
*	$Log: ir.c $
*	 \main\9 9/21/01 3:55 PM dowdellj
*	Fixed warnings after upgrade to MULTI 2000 V3  
*	
*	 \main\8 7/18/01 2:55 PM dogarm
*	revision control details added
*	
*	 \main\7 4/5/01 10:03 AM vickeryr
*	 \main\6 2/16/01 1:13 PM vickeryr
*	 \main\5 2/13/01 10:04 AM vickeryr
*	Added #if MMAC_IR_USED around entire file
*	
*	 \main\4 1/26/01 3:29 PM vickeryr
*	 \main\3 1/24/01 5:13 PM vickeryr
*	Made Capture/Compare Timer used by IR Driver configurable via a #define in mmac_config.h
*	
*	 \main\2 1/11/01 5:04 PM vickeryr
*	 \main\1 1/2/01 2:01 PM vickeryr
*	Made controlled
*	
*
*****************************************************************
*
* Copyright 2001 NEC Electronics (UK) Ltd. 
*
* This file is the property of NEC Electronics (UK) Ltd. The 
* unauthorised use, copy and transfer of all or any part of this file
* is forbidden.
*
****************************************************************/

/****************************************************************/
/*			INCLUDE FILES				*/
/****************************************************************/

#include <linux/module.h>
/* include stdio.h for definition of NULL */
// #include <stdio.h> 
#define NULL ((void *)0)
#include "mmac/types.h"
#include "mmac/assert.h"
#include "mmac/rtos.h"
#include "mmac/capcom.h"
#include "mmac/timer.h"
#include "mmac/ir.h"

/* Only define the IR driver if dictated by mmac_config.h */
#if MMAC_IR_USED

/****************************************************************/
/*			MACROS/DEFINES				*/
/****************************************************************/

/* Set up filename for ASSERT macro */
#if MMAC_DEBUG_ENABLE
static const char assertFileName[] = __FILE__;
#endif

/****************************************************************/
/*			LOCAL TYPES				*/
/****************************************************************/


/****************************************************************/
/*			VARIABLES				*/
/****************************************************************/

/* Subsystem initialisation flag */
#if MMAC_IR_PARAMETER_CHECKING
static MMAC_BOOL infraRedInitialised 	= MMAC_FALSE;
/* Subsystem open flag */
static MMAC_BOOL infraRedOpen 		= MMAC_FALSE;
#endif

/* The Infra Red FIFO - size depends upon value from mmac_config.h       */
/* there is one entry for each transiton read, so buffer size denotes    */
/* the number of transitions which can be held, not the number of bits...*/
static MMAC_IR_READ_DATA infraRedFIFO[MMAC_IR_FIFO_SIZE];

/* Infra Red FIFO buffer 'pointers' */
static UI8 buffNextRead  = 0;
static UI8 buffNextWrite = 0;

/* Structure used to store the current infra red driver configuration */
static MMAC_IR_CONTROL_IOPB 	infraRedConfiguration;

/* value used to store which of the four transitions reported by CAPCOM driver */
/* we should begin reading from, in case we previously timed-out and manually  */
/* read some of them already... */
static UI8	irTransitionToReadNext = 0;

/* UI8 used to store the current subsystem status... */
static MMAC_IR_STATUS	infraRedStatus = 0;

/* MMAC_BOOL used to store the current signal state... */
static MMAC_BOOL infraRedState = MMAC_FALSE;

/* config for TIMER driver used to reset the timeout timer... */
static MMAC_TIMER_WRITE_IOPB 	irTimerReset;

/****************************************************************/
/*			FORWARD DECLARATIONS			*/
/****************************************************************/

/* The Infra Red driver's callback function for the */
/* Capture/Compare driver...                        */
static void IR_CapcomCallback( MMAC_CAPCOM_DEV_NUM dev, MMAC_CAPCOM_PERIOD period );

/* The Infra Red driver's callback function for the */
/* Timer driver...                                  */
static void IR_TimerCallback( MMAC_TIMER_DEV_NUM dev );

/****************************************************************/
/*			FUNCTIONS				*/
/****************************************************************/

/*****************************************************************
*
* Function Name
*	MMAC_IR_Initialise
* 
* Inputs
*	None
*	
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_FAIL
*	MMAC_IR_ALREADY_INITIALISED
*	
* Description
*	This function is used to initialise the Infra Red subsystem.
*	It allocates the resources necessary to manage the IR subsystem.
*	It must be called prior to calling any other function in the
*	Infra Red API.
*	
*	If the subsystem has already been initialised, the function 
*	returns MMAC_IR_ALREADY_INITIALISED without performing any 
*	actions.
*	
*	If the necessary resources cannot be allocated (for the hardware access
*       semaphore, etc), the function will return MMAC_IR_FAIL without
*	initialising the subsystem.
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_Initialise ( void ) 
{
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_TRUE)
    {
	return MMAC_IR_ALREADY_INITIALISED;
    }
#endif
    
    /* Set up default values for the subsystem configuration... */
    infraRedConfiguration.clockPeriod 	 = MMAC_IR_DEFAULT_CLOCK_PERIOD;
    infraRedConfiguration.clockTolerance = MMAC_IR_DEFAULT_CLOCK_TOLERANCE;
    infraRedConfiguration.dataTimeout	 = MMAC_IR_DEFAULT_DATA_TIMEOUT;
    infraRedConfiguration.initialState	 = MMAC_IR_DEFAULT_INITIAL_STATE;
    infraRedConfiguration.callback	 = NULL;
    /* Initialise the status and state... */
    infraRedStatus = MMAC_IR_RX_RESYNC;
    infraRedState  = MMAC_IR_DEFAULT_INITIAL_STATE;
#if MMAC_IR_PARAMETER_CHECKING
    /* Flag subsystem as closed... */
    infraRedOpen = MMAC_FALSE;
    /* Flag the subsystem as being initialised */
    infraRedInitialised = MMAC_TRUE;
#endif

    return MMAC_IR_OK;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_SetConfig
* 
* Inputs
*	iopb 	The configuration to apply to the IR driver:
*
*		clockPeriod	The nominal clock pulse width in
*				microseconds
*		clockTolerance	The tolerance allowed on a single
*				clock pulse in microseconds
*		dataTimeout	The time after which inactivity is
*				deemed to represent an interval
*				between frames
*		callback	The address of the callback function
*				to be called on each data transition
*		initialState	The initial state of the input line
*				(either MMAC_TRUE or MMAC_FALSE)
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_INVALID_PARAM
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_SetConfig ( MMAC_IR_CONTROL_IOPB *iopb ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
    if( iopb == NULL )
    {
	return MMAC_IR_INVALID_PARAM;
    }
    /* Validate fields of iopb... */
    if( iopb->clockPeriod < iopb->clockTolerance )
    {
	return MMAC_IR_INVALID_PARAM;
    }
    if( iopb->dataTimeout < (iopb->clockPeriod + iopb->clockTolerance) )
    {
	return MMAC_IR_INVALID_PARAM;
    }
#endif

    /* Set up the subsystem configuration... */
    infraRedConfiguration.clockPeriod 	 = iopb->clockPeriod;
    infraRedConfiguration.clockTolerance = iopb->clockTolerance;
    infraRedConfiguration.dataTimeout	 = iopb->dataTimeout;
    infraRedConfiguration.initialState	 = iopb->initialState;
    infraRedConfiguration.callback 	 	 = iopb->callback;
    return result;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_GetConfig
* 
* Inputs
*	iopb 	The configuration to apply to the IR driver:
*
*		clockPeriod	The nominal clock pulse width in
*				microseconds
*		clockTolerance	The tolerance allowed on a single
*				clock pulse in microseconds
*		dataTimeout	The time after which inactivity is
*				deemed to represent an interval
*				between frames
*		callback	The address of the callback function
*				to be called on each data transition
*		initialState	The initial state of the input line
*				(either MMAC_TRUE or MMAC_FALSE)
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_INVALID_PARAM
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_GetConfig ( MMAC_IR_CONTROL_IOPB *iopb ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;

    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
    if( iopb == NULL )
    {
	return MMAC_IR_INVALID_PARAM;
    }
#endif

    /* Report the subsystem configuration... */
    iopb->clockPeriod 	 = infraRedConfiguration.clockPeriod;
    iopb->clockTolerance = infraRedConfiguration.clockTolerance;
    iopb->dataTimeout    = infraRedConfiguration.dataTimeout;
    iopb->initialState   = infraRedConfiguration.initialState;
    iopb->callback       = infraRedConfiguration.callback;

    return result;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_Open
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_FAIL
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_ALREADY_OPEN
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_Open ( void ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;
    MMAC_CAPCOM_OPEN_IOPB	irInputConfig;
    MMAC_TIMER_OPEN_IOPB	irTimerConfig;
	MMAC_CAPCOM_CLOSE_IOPB	irInputCloseIopb;
	MMAC_CAPCOM_RESULT_CODE	capComResult;
	MMAC_TIMER_RESULT_CODE	timerResult;
    
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
    if( infraRedOpen == MMAC_TRUE )
    {
	return MMAC_IR_ALREADY_OPEN;
    }
#endif

    /* Set up the capture/compare timer for IR input*/
    /* Nb. This is used to capture transitions...   */
    irInputConfig.deviceNumber 	= (MMAC_CAPCOM_DEV_NUM)MMAC_IR_CAPCOM_TIMER;
    irInputConfig.type		= MMAC_CAPCOM_INPUT;
    irInputConfig.mode		= MMAC_CAPCOM_RESTART;
    irInputConfig.triggerMode	= MMAC_CAPCOM_TRIGGER_BOTH_EDGE;
    irInputConfig.callback	= (MMAC_CAPCOM_CALLBACK)IR_CapcomCallback;
    
    /* Set up the timeout timer config... */
    irTimerConfig.deviceNumber 	= MMAC_TIMER_0;
    irTimerConfig.timeout	= infraRedConfiguration.dataTimeout;
    irTimerConfig.mode		= MMAC_TIMER_SINGLE_SHOT;
    irTimerConfig.callback	= (MMAC_TIMER_CALLBACK)IR_TimerCallback;

    /* Set up the timeout timer reset config... */
    irTimerReset.deviceNumber 	= MMAC_TIMER_0;
    irTimerReset.timeout	= infraRedConfiguration.dataTimeout;

    /* Reset FIFO buffer 'pointers' */
    buffNextRead  = 0;
    buffNextWrite = 0;
   
    /* Flag a re-sync */
    infraRedStatus = MMAC_IR_RX_RESYNC;

    /* Set initial state... */
    infraRedState = infraRedConfiguration.initialState;

    /* Reset next transition to read indicator.... */
    irTransitionToReadNext = 0;

    /* Open the capture/compare input timer... */
    capComResult = MMAC_CAPCOM_Open( &irInputConfig );
	if( capComResult == MMAC_CAPCOM_OK )
	{
    	/* Open the timeout timer... */
	    timerResult = MMAC_TIMER_Open( &irTimerConfig );
		if( timerResult != MMAC_TIMER_OK )
		{
			irInputCloseIopb.deviceNumber = irInputConfig.deviceNumber;
			/* Release the Capture/Cmopare timer... */
			MMAC_CAPCOM_Close( &irInputCloseIopb );
			result = MMAC_IR_FAIL;
		}
    }
	else
	{
		result = MMAC_IR_FAIL;
	}

	if( result == MMAC_IR_OK )
	{
#if MMAC_IR_PARAMETER_CHECKING
    	/* Flag subsystem as open... */
	    infraRedOpen = MMAC_TRUE;
#endif
	}
    
    return result;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_Close
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_Close ( void ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;
    MMAC_CAPCOM_CLOSE_IOPB	capcomIopb;
    MMAC_TIMER_CLOSE_IOPB	timerIopb;
    
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
    if( infraRedOpen == MMAC_FALSE )
    {
	return MMAC_IR_NOT_OPEN;
    }
#endif

    /* Stop the CAPCOM timer...  */
    capcomIopb.deviceNumber = (MMAC_CAPCOM_DEV_NUM)MMAC_IR_CAPCOM_TIMER; 
    MMAC_CAPCOM_Close( &capcomIopb );
    
    /* Stop the timeout timer... */
    timerIopb.deviceNumber = MMAC_TIMER_0; 
    MMAC_TIMER_Close( &timerIopb );
    
#if MMAC_IR_PARAMETER_CHECKING
    /* Flag subsystem as closed... */
    infraRedOpen = MMAC_FALSE;
#endif
    return result;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_Read
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NO_DATA
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	MMAC_IR_INVALID_PARAM
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_Read ( MMAC_IR_READ_DATA *iodata ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;
   
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
    if( infraRedOpen == MMAC_FALSE )
    {
	return MMAC_IR_NOT_OPEN;
    }
    if( iodata == NULL )
    {
	return MMAC_IR_INVALID_PARAM;
    }
#endif

    /* Check FIFO is not empty */
    if( buffNextRead != buffNextWrite )
    {
	/* Return next record from FIFO... */
	iodata->state  = infraRedFIFO[buffNextRead].state;
	iodata->bits   = infraRedFIFO[buffNextRead].bits;
	iodata->status = infraRedFIFO[buffNextRead].status & infraRedStatus;

	/* Now advance the buffNextRead position... */
	buffNextRead++;
	/* Check for FIFO wrap... */
	if( buffNextRead == MMAC_IR_FIFO_SIZE )
	{
	    buffNextRead = 0;
	}
    }
    else
    {
	/* FIFO is empty */
	result = MMAC_IR_NO_DATA;
    }
    return result;
}

/*****************************************************************
*
* Function Name
*	MMAC_IR_Resync
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	
* Description
*	
*****************************************************************/

MMAC_IR_RESULT_CODE MMAC_IR_Resync ( void ) 
{
    MMAC_IR_RESULT_CODE		result = MMAC_IR_OK;
    
    
#if MMAC_IR_PARAMETER_CHECKING
    if( infraRedInitialised == MMAC_FALSE )
    {
	return MMAC_IR_NOT_INITIALISED;
    }
#endif

    /* Clear FIFO... */
    buffNextRead  = 0;
    buffNextWrite = 0;

    /* Flag a re-sync */
    infraRedStatus = MMAC_IR_RX_RESYNC;
    
    return result;
}

/*****************************************************************
*
* Function Name
*	IR_CapcomCallback
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	
* Description
*	Callback function for input capture/compare timer.
*	This function is called each time there have been four
*	transitions on the IR input.
*	When this occurs, the function calculates the number of
*	bits that have been received (based on the period between
*	transitions), and writes the appropriate number of entries
*	into the FIFO buffer.
*
*****************************************************************/

static void IR_CapcomCallback ( MMAC_CAPCOM_DEV_NUM dev, MMAC_CAPCOM_PERIOD period ) 
{
    UI8				transition;
    UI8				numberOfBits;
    
    
    /* If required, clear the timeout bit of infraRedStatus and
       set the resync bit...                                    */
    if( infraRedStatus & MMAC_IR_RX_TIMEOUT )
    {
	infraRedStatus &= ~MMAC_IR_RX_TIMEOUT;
	infraRedStatus |= MMAC_IR_RX_RESYNC;
    }
    
    /* Period contains up to 4 transitons - however, some of these may have    */
    /* already been read by calling MMAC_CAPCOM_Read(). This would be the case */
    /* if data finished with two odd transitions which left the CAPCOM driver  */
    /* waiting for two further transitions before notifying us. We then time   */
    /* out, and maually read the two transitions.                              */
    /* When transitions begin again, they fill transitions 3 and 4 in the      */
    /* 'period' field, and then we are notified of 4 transitions (including    */
    /* the two we have already read).                                          */
    /* Therefore, we need to keep track of which transitions of the four we    */
    /* still need to read...                                                   */
    /* Read any previously un-read transitions from 'period'... */
    for( transition=irTransitionToReadNext; transition<MMAC_CAPCOM_NUM_PERIOD; transition++ )
    {
		/* Only record a number of bits if this is not a resync leading edge... */
		if( infraRedStatus & MMAC_IR_RX_RESYNC )
		{
	    	numberOfBits = 0;
		}
		else
		{
		    numberOfBits = period.times[transition] / infraRedConfiguration.clockPeriod;
		}

		infraRedFIFO[buffNextWrite].state  = infraRedState;
		infraRedFIFO[buffNextWrite].bits   = numberOfBits;
	
		/* Record all apart from overflow in the record entry */
		/* Overflow gets reported if it is occurring when the user calls MMAC_IR_Read() */
		infraRedFIFO[buffNextWrite].status = infraRedStatus & ~MMAC_IR_BUFFER_OVERFLOW;
	
		/* Toggle infraRedState... */
		if( infraRedState == MMAC_TRUE )
		{
		    infraRedState = MMAC_FALSE;
		}
		else
		{
		    infraRedState = MMAC_TRUE;
		}
	
		/* Increment nextWrite index, wrapping if necessary... */
		buffNextWrite++;
		if( buffNextWrite == MMAC_IR_FIFO_SIZE )
		{
		    buffNextWrite = 0;
		}
		/* Check for buffer overflow condition... */
		if( buffNextWrite == buffNextRead)
		{
	    	infraRedStatus |= MMAC_IR_BUFFER_OVERFLOW;
		    /* increment buffNextRead so 'pointers' don't overlap */
		    buffNextRead++;
		    if( buffNextRead == MMAC_IR_FIFO_SIZE )
	    	{
			buffNextRead = 0;
		    }
		}
		else
		{
			/* Clear any overflow condition */
		    infraRedStatus &= ~MMAC_IR_BUFFER_OVERFLOW;
		}

		/* Clear resync now record has been written... */
		infraRedStatus &= ~MMAC_IR_RX_RESYNC;
    }
       
    /* Call any installed callback function... */
    if( infraRedConfiguration.callback != NULL )
    {
	/* Tell callback funtion we have written 'n' more entries in the FIFO... */
	infraRedConfiguration.callback( (UI8)(MMAC_CAPCOM_NUM_PERIOD - irTransitionToReadNext) );
    }

    /* Reset transition to read next... */
    irTransitionToReadNext = 0;
    
    /* Restart the timeout timer... */
    MMAC_TIMER_Write( &irTimerReset );   
}

/*****************************************************************
*
* Function Name
*	IR_TimerCallback
* 
* Inputs
*	None
*
* Outputs
*	None
*	
* Return Codes
*	MMAC_IR_OK
*	MMAC_IR_NOT_INITIALISED
*	MMAC_IR_NOT_OPEN
*	
* Description
*	Callback function for timeout timer.
*	This function is called when there has been no transitions for
*	infraRedConfiguration.dataTimeout milliseconds.
*	When this occurs, the function calls MMAC_CAPCOM_Read(), to
* 	obtain any as yet unreported transitions (the CAPCOM driver will
*	normally wait until it has four transitions to report).
*	The number of outstadning transitions is remembered, so that when
*	the next interrupt from the CAPCOM driver occurs, the first 'n'
*	transitions out of the four reported can be ignored (as they have
*	already been read).
*	Once any outstanding transitions are read, the IR status is set to
*	resync and the timeout timer is not restarted.
*
*****************************************************************/

static void IR_TimerCallback ( MMAC_TIMER_DEV_NUM dev ) 
{
    UI8				numberOfBits;
    UI8				transition;
    UI8				numTransitionsToRead = 0;
    MMAC_CAPCOM_READ_IOPB	readData;
      
    
    /* Manually read any transitions that have not been reported... */
    /* (Could be up to three of these...)                           */
    readData.deviceNumber = (MMAC_CAPCOM_DEV_NUM)MMAC_IR_CAPCOM_TIMER;
    MMAC_CAPCOM_Read( &readData );
    numTransitionsToRead = readData.transitions;
    
    for( transition=0; transition<numTransitionsToRead; transition++ )
    {
		numberOfBits = readData.period.times[transition] / infraRedConfiguration.clockPeriod;
		infraRedFIFO[buffNextWrite].state  = infraRedState;
		infraRedFIFO[buffNextWrite].bits   = numberOfBits;
		
		/* Record all apart from overflow in the record entry */
		/* Overflow gets reported if it is occurring when the user calls MMAC_IR_Read() */
		infraRedFIFO[buffNextWrite].status = infraRedStatus & ~MMAC_IR_BUFFER_OVERFLOW;;
	
		/* Toggle infraRedState... */
		if( infraRedState == MMAC_TRUE )
		{
	    	infraRedState = MMAC_FALSE;
		}
		else
		{
		    infraRedState = MMAC_TRUE;
		}
	
		/* Increment nextWrite index, wrapping if necessary... */
		buffNextWrite++;
		if( buffNextWrite == MMAC_IR_FIFO_SIZE )
		{
		    buffNextWrite = 0;
		}
		/* Check for buffer overflow condition... */
		if( buffNextWrite == buffNextRead)
		{
		    infraRedStatus |= MMAC_IR_BUFFER_OVERFLOW;
		    /* increment buffNextRead so 'pointers' don't overlap */
	    	buffNextRead++;
		    if( buffNextRead == MMAC_IR_FIFO_SIZE )
		    {
			buffNextRead = 0;
		    }
		}
		else
		{
	    	infraRedStatus &= ~MMAC_IR_BUFFER_OVERFLOW;
		}
		
		if( infraRedStatus & MMAC_IR_RX_TIMEOUT )
		{
		    /* Clear resync now record has been written... */
	    	infraRedStatus &= ~MMAC_IR_RX_RESYNC;
		}
    }
    
    /* Write a timeout entry... */
    infraRedFIFO[buffNextWrite].status |= infraRedStatus & MMAC_IR_RX_TIMEOUT;
    infraRedFIFO[buffNextWrite].bits = 0;
    /* Increment nextWrite index, wrapping if necessary... */
    buffNextWrite++;
    if( buffNextWrite == MMAC_IR_FIFO_SIZE )
    {
		buffNextWrite = 0;
    }
	/* Check for buffer overflow condition... */
	if( buffNextWrite == buffNextRead)
	{
	    infraRedStatus |= MMAC_IR_BUFFER_OVERFLOW;
	    /* increment buffNextRead so 'pointers' don't overlap */
	   	buffNextRead++;
	    if( buffNextRead == MMAC_IR_FIFO_SIZE )
	    {
			buffNextRead = 0;
		}
	}
	else
	{
	   	infraRedStatus &= ~MMAC_IR_BUFFER_OVERFLOW;
	}
    
    /* Remember the next valid transition's position... */
    irTransitionToReadNext = numTransitionsToRead;

    /* Call any installed callback function... */
    if( infraRedConfiguration.callback != NULL )
    {
		/* Tell callback function we have written 'n' more entries in the FIFO... */
		infraRedConfiguration.callback( numTransitionsToRead );
    }

    /* Flag a timeout */
    infraRedStatus |= MMAC_IR_RX_TIMEOUT;   
}

/****************************************************************/
/*			END OF UNIT				*/
/****************************************************************/
MODULE_DESCRIPTION("ir module");
MODULE_AUTHOR("NEC");
#endif
